from scrilla.gui.widgets.functions import CorrelationWidget, DiscountDividendWidget, DistributionWidget, \
    EfficientFrontierWidget, MovingAverageWidget, OptimizerWidget, \
    RiskProfileWidget, YieldCurveWidget

FUNC_WIDGETS = {
    'correlation': {
        'name': 'Correlation Matrix',
        'class': CorrelationWidget,
        'shortcut': 'Ctrl+1',
        'group': 'analysis'
    },
    'dividend': {
        'name': 'Discount Dividend Model',
        'class': DiscountDividendWidget,
        'shortcut': 'Ctrl+2',
        'group': 'analysis'
    },
    'distribution': {
        'name': 'Distribution of Returns',
        'class': DistributionWidget,
        'shortcut': 'Ctrl+7',
        'group': 'analysis'
    },
    'frontier': {
        'name': 'Efficient Frontiers',
        'class': EfficientFrontierWidget,
        'shortcut': 'Ctrl+3',
        'group': 'analysis'
    },
    'averages': {
        'name': 'Moving Averages',
        'class': MovingAverageWidget,
        'shortcut': 'Ctrl+4',
        'group': 'analysis'
    },
    'optimize': {
        'name': 'Portfolio Optimization',
        'class': OptimizerWidget,
        'shortcut': 'Ctrl+5',
        'group': 'allocation'
    },
    'risk_profile': {
        'name': 'Risk Profile',
        'class': RiskProfileWidget,
        'shortcut': 'Ctrl+8',
        'group': 'analysis'

    },
    'yield_curve': {
        'name': 'Yield Curve',
        'class': YieldCurveWidget,
        'shortcut': 'Ctrl+9',
        'group': 'prices'
    }
}

MENUBAR_WIDGET = {
    'Account': [{
        'name': 'Add API Key',
        'shortcut': 'Ctrl+A',
    }],
    'View': [{
        'name': 'Function Menu',
        'shortcut': 'Ctrl+F'
    },
        {
        'name': 'Splash Menu',
        'shortcut': 'Ctrl+S'
    }],
    'Functions': [{
        'name': FUNC_WIDGETS[func_widget]['name'],
        'shortcut': FUNC_WIDGETS[func_widget]['shortcut']
    } for func_widget in FUNC_WIDGETS
    ],
}



# This file is part of scrilla: https://github.com/chinchalinchin/scrilla.

# scrilla is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3
# as published by the Free Software Foundation.

# scrilla is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with scrilla.  If not, see <https://www.gnu.org/licenses/>
# or <https://github.com/chinchalinchin/scrilla/blob/develop/main/LICENSE>.

import datetime

from scrilla import services, settings, errors
from scrilla.static import constants
from scrilla.util import dater, outputter
import scrilla.analysis.estimators as estimators

logger = outputter.Logger(
    'scrilla.analysis.objects.cashflow', settings.LOG_LEVEL)

# Technically these are periods
FREQ_DAY = 1/365
FREQ_MONTH = 1/12
FREQ_QUARTER = 1/4
FREQ_ANNUAL = 1

# Frequency = 1 / period


class Cashflow:
    r"""
    A class that represents a set of future cashflows. The class is initialized with the `sample` variable, a `dict` of past cashflows and their dates. From the `sample`, a linear regression model is inferred. Alternatively, a `growth_function` can be provided that describes the cash flow as a function of time measured in years. If a `growth_function` is provided, the class skips the linear regression model. See warning below for more information on constructing an instance of this cashflow. In general, it needs to know how to project future cashflows, whether that is through inference or a model assumption. 

    If the sample of data is not large enough to infer a linear regression model, the estimation model will default to simple Martingale process described by the equation,

    $$ E(X_2 \mid X_1) = X_1 $$

    Or, in plain English, the next expected value given the current value is the current value. To rephrase it yet again: without more information the best guess for the future value of an asset is its current value.

    The growth model, whether estimated or provided, is used to project the future value of cashflows and then these projections are discounted back to the present by the `discount_rate`. 

    Parameters
    ----------
    1. **sample**: ``list``
        *Optional*. A list comprised of the cashflow\'s historical values. The list must be ordered from latest to earliest, i.e. in descending order. Must be of the format: `{ 'date_1' : 'value_1', 'date_2': 'value_2', ... }`
    2. **period**: ``float``
        *Optional*. The period between the cash flow payments. Measured as the length of time between two distinct cash flows, assuming all such payments are evenly spaced across time. The value should be measured in years. If a period is not specified, then a period will be inferred from the sample of data by averaging the time periods between successive payments in the sample. Common period are statically accessible through `FREQ_DAY`, `FREQ_MONTH`, `FREQ_QUARTER` and `FREQ_ANNUAL`. (*Yes, I know period = 1 / frequency; deal with it.*) 
    3. **growth_function**: ``function``
        *Optional*. A function that describes the cash flow as a function of time in years. If provided, the class will skip linear regression for estimating the cash flow model. If a `growth_function` is provided without a sample, a period must be specified. If a `growth_function` is provided with a sample and no period, the period will be inferred from the dates in the sample. If a `growth_function` is provided with a period, then the sample will be ignored altogether.
    4. **discount_rate**: ``float``
        *Optional.* The rate of return used to discount future cash flows back to the present. If not provided, the `discount_rate` defaults to the risk free rate defined by the **RISK_FREE** environment variable.
    5. **constant**: ``float``
        If the cashflow is constant with respect to time, specify the value of it with this argument. Will override `growth_function` and sample. If constant is specified, you MUST also specify a period or else you will encounter errors when trying to calculate the net present value of future cashflows.

    .. warning::
    * In general, the Cashflow object must always be initialized in one of the following ways:
        1. **__init__** args: (`sample`) -> period inferred from sample, linear regression used for growth
        2. **__init__** args: (`sample`, `period`) -> period from constructor, linear regression used for growth
        3. **__init__** args: (`sample`, `period`, `growth_function`) -> period from constructor, `growth_function` used for growth, sample ignored
        4. **__init__** args: (`sample`, `growth_function`) -> period inferred from sample, `growth_function` used for growth
        5. **__init__** args: (`period`, `growth_function`) -> period from constructor, `growth_function` used for growth
        6. **__init__** args: (`period`, `constant`) -> period from constructor, constant used for growth

    .. notes::
        * A constant cashflow can be specified in three ways: 1. By passing in a constant amount through the constructor `constant` variable. 2. By passing in a constant function with respect to time through the constructor `growth_function` variable. 3. By passing in a dataset of length one through the constructor `sample` variable.  In any of the cases, you MUST pass in a period or the `net_present_value` method of this class will return False.
        * Both a growth_function and a sample of data can be passed in at once to this class. If doing so, the `growth_function` will take precedence and be used for calculations in the `net_present_value` method. The sample will be used to infer the length of a period between cashflows, unless a period is also specified. If a period is specified in addition to `sample_prices` and `growth_function`, the period will take precedence over the period inferred from the sample of data.
        * The coefficients of the inferred linear regression are accessibly through `Cashflow().alpha` (intercept) and `Cashflow().beta` (slope) instance variables. The time series used to create the model is accessible through the `Cashflow().time_series` instance variable; Note: it is measured in years and the `start_date` is set equal to 0. In other words, the intercept of the model represents, approximately, the value of the cashflow on the `start_date`.

    .. todos::
        * Implement prediction interval function to get error bars for graphs and general usage.

    """

    def __init__(self, sample=None, period=None, growth_function=None, constant=None, discount_rate=None):
        self.sample = sample
        self.period = period
        self.growth_function = growth_function

        # if constant is specified, override sample and growth_function
        if constant is not None:
            logger.debug(f'constant = $ {constant}; period MUST NOT be null!')
            logger.debug(f'period = {self.period}')
            self.constant = constant
            self.sample = None
            self.growth_function = None
        else:
            self.constant = None

        # If sample provided, use simple linear regression
        if self.sample is not None and self.growth_function is None:
            self.generate_time_series_for_sample()
            self.regress_growth_function()

        if discount_rate is None:
            self.discount_rate = services.get_risk_free_rate()
        else:
            self.discount_rate = discount_rate

        logger.debug(f'Using discount_rate = {self.discount_rate}')

        # If no frequency is specified, infer frequency from sample
        if self.sample is not None and self.period is None:
            self.infer_period()

        if self.sample is not None and len(self.sample) > 0:
            self.time_to_today = self.calculate_time_to_today()

    def infer_period(self):
        logger.debug('Attempting to infer period/frequency of cashflows.')

        # no_of_dates = len - 1 because delta is being computed, i.e.
        #   lose one date.
        dates, no_of_dates = self.sample.keys(), (len(self.sample.keys()) - 1)
        first_pass = True
        mean_delta = 0

        if no_of_dates < 2:
            logger.debug(
                'Cannot infer period from sample size less than or equal to 1')
            self.period = None
            self.frequency = None

        else:
            for date in dates:
                if first_pass:
                    tomorrows_date = dater.parse(date)
                    first_pass = False

                else:
                    todays_date = dater.parse(date)
                    delta = (tomorrows_date - todays_date).days / 365
                    mean_delta += delta / no_of_dates
                    tomorrows_date = todays_date

            self.period = mean_delta
            self.frequency = 1 / self.period
            logger.debug(f'Inferred period = {self.period} yrs')
            logger.debug(f'Inferred frequency = {self.frequency}')

    # TODO: trading days or actual days?
    def generate_time_series_for_sample(self):
        self.time_series = []

        dates, no_of_dates = self.sample.keys(), len(self.sample.keys())

        if no_of_dates == 0:
            logger.debug(
                'Cannot generate a time series for a sample size of 0.')
            self.time_series = None
        else:
            first_date = dater.parse(list(dates)[-1])

            for date in dates:
                this_date = dater.parse(date)
                delta = (this_date - first_date).days
                time_in_years = delta / 365
                self.time_series.append(time_in_years)

    # TODO: trading days or actual days?
    def calculate_time_to_today(self):
        first_date = dater.parse(list(self.sample.keys())[-1])
        today = datetime.date.today()
        return ((today - first_date).days/365)

    def regress_growth_function(self):
        to_list = [self.sample[date] for date in self.sample]

        self.beta = estimators.simple_regression_beta(
            x=self.time_series, y=to_list)
        self.alpha = estimators.simple_regression_alpha(
            x=self.time_series, y=to_list)

        if not self.beta or not self.alpha:
            if len(self.sample) > 0:
                self.alpha = list(self.sample.items())[0][1]
                logger.debug(
                    'Error calculating regression coefficients; Defaulting to Markovian process E(X2|X1) = X1.')
                logger.debug(f'Estimation model : y = {self.alpha}')
            else:
                raise errors.SampleSizeError(
                    'Not enough information to formulate estimation model.')

        else:
            logger.debug(
                f'Linear regression model : y = {self.beta} * x + {self.alpha}')

    def generate_model_series(self):
        return [self.alpha + self.beta*time for time in self.time_series]

    def generate_model_comparison(self):
        """
        Returns a list of dictionaries with the predicted value of the linear regression model and the actual value on a given datas. Format: [ {'date': `str`, 'model_price': `float`, 'actual_price': `float` }, ... ]
        """
        model_prices = self.generate_model_series()

        return[{'date': date,
                'model_price': model_prices[index],
                'actual_price': self.sample[date]}
               for index, date in enumerate(self.sample.keys())]

    def get_growth_function(self, x):
        """
        Traverses the hierarchy of instance variables to determine which method to use to describe the growth of future cashflows. Returns the value of determined function for the given value of `x`. Think of this function as a black box that hides the implementation of the `growth_function` from the user accessing the function. 

        Parameters
        ----------
        1. **x**: ``float``
            Time in years.

        Returns
        -------
        ``float`` : Value of the cash flow's growth function at time `x`.

        """
        if self.growth_function is None:
            if self.constant is not None:
                return self.constant
            return (self.alpha + self.beta*(x + self.time_to_today))
        return self.growth_function(x)

    # TODO: use trading days or actual days?
    def calculate_net_present_value(self):
        """
        Returns the net present value of the cash flow by using the `get_growth_function` method to project future cash flows and then discounting those projections back to the present by the value of the `discount_rate`. Call this method after constructing/initializing a `Cashflow` object to retrieve its NPV.

        Raises
        ------
        1. **scrilla.errors.InputValidationError**
            If not enough information is present in the instance of the `Cashflow` object to project future cash flows, this error will be thrown.

        Returns
        -------
        ``float`` : NPV of cash flow.
        """
        if self.period is None:
            raise errors.InputValidationError(
                "No period detected for cashflows. Not enough information to calculate net present value.")

        time_to_first_payment = 0
        if self.period is None:
            raise errors.InputValidationError(
                'Not enough information to calculate net present value of cash flow.')
        if self.period == FREQ_ANNUAL:
            time_to_first_payment = dater.get_time_to_next_year()

        elif self.period == FREQ_QUARTER:
            time_to_first_payment = dater.get_time_to_next_quarter()

        elif self.period == FREQ_MONTH:
            time_to_first_payment = dater.get_time_to_next_month()

        elif self.period == FREQ_DAY:
            time_to_first_payment = FREQ_DAY

        else:
            dates = self.sample.keys()
            latest_date = dater.parse(list(dates)[0])
            time_to_first_payment = dater.get_time_to_next_period(
                starting_date=latest_date, period=self.period)

        net_present_value, i, current_time = 0, 0, 0
        calculating = True
        while calculating:
            previous_value = net_present_value
            current_time = time_to_first_payment + i * self.period

            net_present_value += self.get_growth_function(current_time) / (
                (1 + self.discount_rate)**current_time)

            if net_present_value - previous_value < constants.constants['NPV_DELTA_TOLERANCE']:
                calculating = False
            i += 1

        return net_present_value



# This file is part of scrilla: https://github.com/chinchalinchin/scrilla.

# scrilla is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3
# as published by the Free Software Foundation.

# scrilla is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with scrilla.  If not, see <https://www.gnu.org/licenses/>
# or <https://github.com/chinchalinchin/scrilla/blob/develop/main/LICENSE>.

"""
# CLI Entrypoint 
This script acts as the entrypoint for the CLI application and contains the majority of the control structures for the program. It parses the arguments supplied through the command line, delegates them to the appropriate application function and then passes the results to the `scrilla.util.outputter` module functions for formatting and printing to screen.

The arguments are parsed in such a way that arguments which are not supplied are set to None. All application functions are set up to accept None as a value for their optional arguments. This makes passing arguments to application functions easier as the `main.py` script doesn't have to worry about their values. In other words, `main.py` always passes all arguments to application functions, even if they aren't supplied through the command line; it just sets the ones which aren't supplied to None.
"""

import time
from datetime import date
from typing import Callable, Dict, List, Union

from scrilla.settings import LOG_LEVEL, ESTIMATION_METHOD
from scrilla.static import definitions
from scrilla.errors import InputValidationError
from scrilla.files import init_static_data
from scrilla.util.helper import format_args
from scrilla.util.outputter import Logger

# TODO: conditional imports based on value of ANALYSIS_MODE

logger = Logger('main', LOG_LEVEL)


def validate_function_usage(selection: str, args: List[str], wrapper_function: Callable, required_length: int = 1, exact: bool = False) -> None:
    """
    Parameters
    ----------
    1. **selection** : ``str``
    2. **args**: ``list``
        List of ticker/statistic symbols whose length is be to validated.
    3. **wrapper_function** : ``func()``
    4. **required_length** : ``int``
        The number of tickers required for the function.
    5. **exact** : ``bool``
        *Optional*. If the required length constraint is an equality, set to `True`. If the constraint is an inequality, set to `False`. Defaults to `False`. 
    """

    start_time = time.time()
    if(not exact and (len(args) > (required_length-1))) or (exact and (len(args) == required_length)):
        wrapper_function()
    elif exact:
        raise InputValidationError(
            f'Invalid number of arguments for \'{selection}\' function. Function requires {required_length} arguments.')
    else:
        raise InputValidationError(
            f'Invalid number of arguments for \'{selection}\' function. Function requires more than {required_length} arguments.')
    end_time = time.time()
    logger.info(f'Total execution time: {end_time - start_time}s')


def print_format_to_screen(args: Dict[str, Union[str, date, float, None, bool]]) -> bool:
    """
    Checks if the inputted optional arguments allow printing pretty formatted text to screen.

    Parameters
    ----------
    1. **args** : ``dict``
        Formatted dictionary containing optional arguments. Result of a call to `helper.format_xtra_args_dict`.

    Returns
    -------
    ``bool``
    """
    return not args['json'] and not args['suppress_output']


def print_json_to_screen(args: Dict[str, Union[str, date, float, None, bool]]) -> bool:
    """
    Checks if inputted optional arguments allow printing json formatted text to screen.

    Parameters
    ----------
    1. **args** : ``dict``
        Formatted dictionary containing optional arguments. Result of a call to `helper.format_xtra_args_dict`.

    Returns
    -------
    ``bool``
    """
    return args['json'] and not args['suppress_output']


def do_program(cli_args: List[str]) -> None:
    """
    Parses command line arguments and passes the formatted arguments to appropriate function from the library.
    """
    init_static_data()
    args = format_args(cli_args, ESTIMATION_METHOD)
    exact, selected_function = False, None

    # START CLI FUNCTION DEFINITIONS

    # NO ARGUMENT FUNCTIONS
    # FUNCTION: Help Message
    if args['function_arg'] in definitions.FUNC_DICT["help"]['values']:
        def cli_help():
            from scrilla.util.outputter import help_msg
            help_msg()
        selected_function, required_length = cli_help, 0

    # FUNCTION: Clear Cache
    elif args['function_arg'] in definitions.FUNC_DICT["clear_cache"]['values']:
        def cli_clear_cache():
            from scrilla.files import clear_directory
            from scrilla.settings import CACHE_DIR
            logger.comment(f'Clearing {CACHE_DIR}')
            clear_directory(directory=CACHE_DIR, retain=True)
        selected_function, required_length = cli_clear_cache, 0

    # FUNCTION: Clear Static
    elif args['function_arg'] in definitions.FUNC_DICT["clear_static"]['values']:
        def cli_clear_static():
            from scrilla.files import clear_directory
            from scrilla.settings import STATIC_DIR
            logger.comment(f'Clearing {STATIC_DIR}')
            clear_directory(directory=STATIC_DIR, retain=True)
        selected_function, required_length = cli_clear_static, 0

    # FUNCTION: Clear Common
    elif args['function_arg'] in definitions.FUNC_DICT["clear_common"]['values']:
        def cli_clear_common():
            from scrilla.files import clear_directory
            from scrilla.settings import COMMON_DIR
            logger.comment(f'Clearing {COMMON_DIR}')
            clear_directory(directory=COMMON_DIR, retain=True)
        selected_function, required_length = cli_clear_common, 0

    # FUNCTION: Print Stock Watchlist
    elif args['function_arg'] in definitions.FUNC_DICT['list_watchlist']['values']:
        def cli_watchlist():
            from scrilla.files import get_watchlist
            from scrilla.util.outputter import title_line, print_list
            tickers = get_watchlist()
            title_line("Stock Watchlist")
            print_list(tickers)
        selected_function, required_length = cli_watchlist, 0
    # FUNCTION: Purge Data Directories
    elif args['function_arg'] in definitions.FUNC_DICT["purge"]['values']:
        def cli_purge():
            from scrilla.files import clear_directory
            from scrilla.settings import CACHE_DIR, STATIC_DIR, COMMON_DIR
            logger.comment(
                f'Clearing {STATIC_DIR}, {CACHE_DIR} and {COMMON_DIR}')
            clear_directory(directory=STATIC_DIR, retain=True)
            clear_directory(directory=CACHE_DIR, retain=True)
            clear_directory(directory=COMMON_DIR, retain=True)
        selected_function, required_length = cli_purge, 0

    # FUNCTION: Display Version
    elif args['function_arg'] in definitions.FUNC_DICT["version"]['values']:
        def cli_version():
            from scrilla.settings import APP_DIR
            from os.path import join
            version_file = join(APP_DIR, 'version.txt')
            with open(version_file, 'r') as f:
                print(f.read())
        selected_function, required_length = cli_version, 0

    # FUNCTION: Yield Curve
    elif args['function_arg'] in definitions.FUNC_DICT['yield_curve']['values']:
        def cli_yield_curve():
            from scrilla.static.keys import keys
            from scrilla.services import get_daily_interest_history
            yield_curve = {}
            for maturity in keys['YIELD_CURVE']:
                curve_rate = get_daily_interest_history(maturity=maturity,
                                                        start_date=args['start_date'],
                                                        end_date=args['start_date'])
                yield_curve[maturity] = curve_rate[list(
                    curve_rate.keys())[0]]/100

                if print_format_to_screen(args):
                    from scrilla.util.outputter import scalar_result
                    scalar_result(calculation=maturity,
                                  result=yield_curve[maturity], currency=False)

            if print_json_to_screen(args):
                from json import dumps
                print(dumps(yield_curve))

            if args['save_file'] is not None:
                from scrilla.files import save_file
                save_file(file_to_save=yield_curve,
                          file_name=args['save_file'])
        selected_function, required_length = cli_yield_curve, 0

    # ARGUMENT FUNCTIONS
    # FUNCTION: Asset Grouping

    elif args['function_arg'] in definitions.FUNC_DICT['asset_type']['values']:
        def cli_asset_type():
            from scrilla.files import get_asset_type
            from scrilla.util.outputter import string_result
            for arg in args['tickers']:
                asset_type = get_asset_type(arg)
                string_result(f'asset_type({arg})', asset_type)
        selected_function, required_length = cli_asset_type, 1

    # FUNCTION: Black-Scholes Value At Risk
    elif args['function_arg'] in definitions.FUNC_DICT['var']['values']:
        def cli_var():
            from scrilla.analysis.models.geometric.statistics import calculate_risk_return
            from scrilla.analysis.models.geometric.probability import percentile
            from scrilla.services import get_daily_price_history
            from scrilla.util.helper import get_first_json_key
            from scrilla.static.keys import keys

            all_vars = {}
            for arg in args['tickers']['values']:
                prices = get_daily_price_history(ticker=arg,
                                                 start_date=args['start_date'],
                                                 end_date=args['end_date'])
                latest_price = prices[get_first_json_key(
                    prices)][keys['PRICES']['CLOSE']]
                profile = calculate_risk_return(ticker=arg,
                                                sample_prices=prices,
                                                method=args['estimation_method'])
                valueatrisk = percentile(S0=latest_price,
                                         vol=profile['annual_volatility'],
                                         ret=profile['annual_return'],
                                         expiry=args['expiry'],
                                         prob=args['probability'])

                all_vars[arg] = valueatrisk

                if print_format_to_screen(args):
                    from scrilla.util.outputter import scalar_result
                    scalar_result(f'{arg}_VaR', valueatrisk)

            if print_json_to_screen(args):
                from json import dumps
                print(dumps(all_vars))

            if args['save_file'] is not None:
                from scrilla.files import save_file
                save_file(file_to_save=all_vars, file_name=args['save_file'])

        selected_function, required_length = cli_var, 2

    # FUNCTION: Black-Scholes Conditional Value At Risk
    elif args['function_arg'] in definitions.FUNC_DICT['cvar']['values']:
        def cli_cvar():
            from scrilla.services import get_daily_price_history
            from scrilla.static.keys import keys
            from scrilla.analysis.models.geometric.statistics import calculate_risk_return
            from scrilla.analysis.models.geometric.probability import percentile, conditional_expected_value
            from scrilla.util.helper import get_first_json_key
            all_cvars = {}
            for arg in args['tickers']:
                prices = get_daily_price_history(ticker=arg,
                                                 start_date=args['start_date'],
                                                 end_date=args['end_date'])
                latest_price = prices[get_first_json_key(
                    prices)][keys['PRICES']['CLOSE']]
                profile = calculate_risk_return(ticker=arg,
                                                sample_prices=prices,
                                                method=args['estimation_method'])
                valueatrisk = percentile(S0=latest_price,
                                         vol=profile['annual_volatility'],
                                         ret=profile['annual_return'],
                                         expiry=args['expiry'],
                                         prob=args['probability'])
                cvar = conditional_expected_value(S0=latest_price,
                                                  vol=profile['annual_volatility'],
                                                  ret=profile['annual_return'],
                                                  expiry=args['expiry'],
                                                  conditional_value=valueatrisk)
                all_cvars[arg] = cvar

                if print_format_to_screen(args):
                    from scrilla.util.outputter import scalar_result
                    scalar_result(
                        f'{arg}_conditional_value_at_risk', valueatrisk)

            if print_json_to_screen(args):
                from json import dumps
                print(dumps(all_cvars))

            if args['save_file'] is not None:
                from scrilla.files import save_file
                save_file(file_to_save=all_cvars, file_name=args['save_file'])

        selected_function, required_length = cli_cvar, 2

    # FUNCTION: Capital Asset Pricing Model Cost of Equity
    elif args['function_arg'] in definitions.FUNC_DICT['capm_equity_cost']['values']:
        def cli_capm_equity_cost():
            from scrilla.analysis.markets import cost_of_equity
            all_costs = {}
            for arg in args['tickers']:
                equity_cost = cost_of_equity(ticker=arg,
                                             start_date=args['start_date'],
                                             end_date=args['end_date'],
                                             method=args['estimation_method'])
                all_costs[arg] = equity_cost

                if print_format_to_screen(args):
                    from scrilla.util.outputter import scalar_result
                    scalar_result(f'{arg}_equity_cost',
                                  equity_cost, currency=False)

            if print_json_to_screen(args):
                from json import dumps
                print(dumps(all_costs))

            if args['save_file'] is not None:
                from scrilla.files import save_file
                save_file(file_to_save=all_costs, file_name=args['save_file'])

        selected_function, required_length = cli_capm_equity_cost, 1

    # FUNCTION: Capital Asset Pricing Model Beta
    elif args['function_arg'] in definitions.FUNC_DICT['capm_beta']['values']:
        def cli_capm_beta():
            from scrilla.analysis.markets import market_beta
            all_betas = {}
            for arg in args['tickers']:
                beta = market_beta(ticker=arg,
                                   start_date=args['start_date'],
                                   end_date=args['end_date'],
                                   method=args['estimation_method'])
                all_betas[arg] = beta

                if print_format_to_screen(args):
                    from scrilla.util.outputter import scalar_result
                    scalar_result(f'{arg}_beta', beta, currency=False)

            if print_json_to_screen(args):
                from json import dumps
                print(dumps(all_betas))

            if args['save_file'] is not None:
                from scrilla.files import save_file
                save_file(file_to_save=all_betas, file_name=args['save_file'])
        selected_function, required_length = cli_capm_beta, 1

    # FUNCTION: Last Close Price
    elif args['function_arg'] in definitions.FUNC_DICT["close"]['values']:
        def cli_close():
            from scrilla.services import get_daily_price_latest

            all_prices = {}
            for arg in args['tickers']:
                price = get_daily_price_latest(arg)
                all_prices[arg] = price

                if print_format_to_screen(args):
                    from scrilla.util.outputter import scalar_result
                    scalar_result(
                        calculation=f'Last {arg} close price', result=float(price))

            if print_json_to_screen(args):
                from json import dumps
                print(dumps(all_prices))

            if args['save_file'] is not None:
                from scrilla.files import save_file
                save_file(file_to_save=all_prices, file_name=args['save_file'])
        selected_function, required_length = cli_close, 1

    # FUNCTION: Correlation Matrix
    elif args['function_arg'] in definitions.FUNC_DICT["correlation"]['values']:
        def cli_correlation():
            from scrilla.static.keys import keys
            from scrilla.analysis.models.geometric.statistics import correlation_matrix

            if args['estimation_method'] == keys['ESTIMATION']['LIKE']:
                logger.comment('This calculation takes a while, strap in...')

            matrix = correlation_matrix(tickers=args['tickers'],
                                        start_date=args['start_date'],
                                        end_date=args['end_date'],
                                        method=args['estimation_method'])

            if print_format_to_screen(args):
                from scrilla.util.outputter import correlation_matrix as correlation_output
                correlation_output(
                    tickers=args['tickers'], correl_matrix=matrix)

            elif print_json_to_screen(args):
                from json import dumps
                print(dumps(matrix))

            if args['save_file'] is not None:
                from scrilla.files import save_file
                save_file(file_to_save=matrix, file_name=args['save_file'])

        selected_function, required_length = cli_correlation, 2

    # FUNCTION: Correlation Time Series
    elif args['function_arg'] in definitions.FUNC_DICT['correlation_time_series']['values']:
        def cli_correlation_series():
            from scrilla.analysis.models.geometric.statistics import calculate_moment_correlation_series
            if print_format_to_screen(args):
                from scrilla.util.outputter import scalar_result

            logger.comment('This calculation takes a while, strap in...')
            ticker_1, ticker_2 = args['tickers'][0], args['tickers'][1]
            result = calculate_moment_correlation_series(ticker_1=ticker_1,
                                                         ticker_2=ticker_2,
                                                         start_date=args['start_date'],
                                                         end_date=args['end_date'])
            if print_format_to_screen(args):
                for this_date in result:
                    scalar_result(calculation=f'{this_date}_{ticker_1}_{ticker_2}_correlation',
                                  result=float(result[this_date]), currency=False)
            elif print_json_to_screen(args):
                from json import dumps
                print(dumps(result))

            if args['save_file'] is not None:
                from scrilla.files import save_file
                save_file(file_to_save=result, file_name=args['save_file'])
        selected_function, required_length, exact = cli_correlation_series, 2, True

    # FUNCTION: Discount Dividend Model
    elif args['function_arg'] in definitions.FUNC_DICT["discount_dividend"]['values']:
        def cli_discount_dividend():
            from scrilla.services import get_dividend_history
            from scrilla.analysis.objects.cashflow import Cashflow
            model_results = {}

            for arg in args['tickers']:
                dividends = get_dividend_history(arg)
                if args['discount'] is None:
                    from scrilla.analysis.markets import cost_of_equity
                    discount = cost_of_equity(
                        ticker=arg, method=args['estimation_method'])
                else:
                    discount = args['discount']
                model_results[f'{arg}_discount_dividend'] = Cashflow(sample=dividends,
                                                                     discount_rate=discount).calculate_net_present_value()
                if print_format_to_screen(args):
                    from scrilla.util.outputter import scalar_result
                    scalar_result(f'Net Present Value ({arg} dividends)',
                                  model_results[f'{arg}_discount_dividend'])

            if print_json_to_screen(args):
                from json import dumps
                print(dumps(model_results))

            if args['save_file'] is not None:
                from scrilla.files import save_file
                save_file(file_to_save=model_results,
                          file_name=args['save_file'])
        selected_function, required_length = cli_discount_dividend, 1

    elif args['function_arg'] in definitions.FUNC_DICT['dividends']['values']:
        def cli_dividends():
            from scrilla.services import get_dividend_history
            if print_format_to_screen(args):
                from scrilla.util.outputter import scalar_result

            all_dividends = {}
            for arg in args['tickers']:
                dividends = get_dividend_history(arg)
                all_dividends[arg] = dividends

                if print_format_to_screen(args):
                    for this_date in dividends:
                        scalar_result(
                            calculation=f'{arg}_dividend({this_date})', result=dividends[this_date])

            if print_json_to_screen(args):
                from json import dumps
                print(dumps(all_dividends))

        selected_function, required_length = cli_dividends, 1

    # FUNCTION: Efficient Frontier
    elif args['function_arg'] in definitions.FUNC_DICT['efficient_frontier']['values']:
        def cli_efficient_frontier():
            from scrilla.analysis.objects.portfolio import Portfolio
            from scrilla.analysis.optimizer import calculate_efficient_frontier

            portfolio = Portfolio(tickers=args['tickers'],
                                  start_date=args['start_date'],
                                  end_date=args['end_date'],
                                  method=args['estimation_method'])
            frontier = calculate_efficient_frontier(portfolio=portfolio,
                                                    steps=args['steps'])

            if args['investment'] is not None:
                from scrilla.services import get_daily_prices_latest
                prices = get_daily_prices_latest(
                    tickers=args['tickers'])
            else:
                prices = None

            if print_format_to_screen(args):
                from scrilla.util.outputter import efficient_frontier as frontier_output
                frontier_output(portfolio=portfolio,
                                frontier=frontier,
                                investment=args['investment'],
                                latest_prices=prices)
            if print_json_to_screen(args):
                from json import dumps
                from scrilla.static.formats import format_frontier
                print(dumps(format_frontier(portfolio=portfolio,
                                            frontier=frontier,
                                            investment=args['investment'],
                                            latest_prices=prices)))

            if args['save_file'] is not None:
                from scrilla.files import save_frontier
                save_frontier(portfolio=portfolio,
                              frontier=frontier,
                              investment=args['investment'],
                              file_name=args['save_file'],
                              latest_prices=prices)
        selected_function, required_length = cli_efficient_frontier, 2

    # FUNCTION: Maximize Portfolio Return
    elif args['function_arg'] in definitions.FUNC_DICT['maximize_return']['values']:
        def cli_maximize_return():
            from scrilla.analysis.objects.portfolio import Portfolio
            from scrilla.analysis.optimizer import maximize_portfolio_return

            portfolio = Portfolio(tickers=args['tickers'],
                                  start_date=args['start_date'],
                                  end_date=args['end_date'],
                                  method=args['estimation_method'])

            allocation = maximize_portfolio_return(
                portfolio=portfolio)

            if args['investment'] is not None:
                from scrilla.services import get_daily_prices_latest
                prices = get_daily_prices_latest(
                    tickers=args['tickers'])
            else:
                prices = None

            if print_format_to_screen(args):
                from scrilla.util.outputter import optimal_result
                optimal_result(portfolio=portfolio,
                               allocation=allocation,
                               investment=args['investment'],
                               latest_prices=prices)

            if print_json_to_screen(args):
                from json import dumps
                from scrilla.static.formats import format_allocation
                print(dumps(format_allocation(allocation=allocation,
                                              portfolio=portfolio,
                                              investment=args['investment'],
                                              latest_prices=prices)))

            if args['save_file'] is not None:
                from scrilla.files import save_allocation
                save_allocation(allocation=allocation,
                                portfolio=portfolio,
                                file_name=args['save_file'],
                                investment=args['investment'],
                                latest_prices=prices)
        selected_function, required_length = cli_maximize_return, 2

    # FUNCTION: Moving Averages of Logarithmic Returns
    elif args['function_arg'] in definitions.FUNC_DICT['moving_averages']['values']:
        def cli_moving_averages():
            from scrilla.analysis.models.geometric.statistics import calculate_moving_averages
            # TODO: moving averages with estimation techniques

            moving_averages = calculate_moving_averages(ticker=args['tickers'][0],
                                                        start_date=args['start_date'],
                                                        end_date=args['end_date'],
                                                        method=args['estimation_method'])

            if print_format_to_screen(args):
                from scrilla.util.outputter import moving_average_result
                moving_average_result(
                    ticker=args['tickers'][0], averages=moving_averages)

            if print_json_to_screen(args):
                from json import dumps
                print(dumps(moving_averages))

            if args['save_file'] is not None:
                from scrilla.files import save_file
                save_file(file_to_save=moving_averages,
                          file_name=args['save_file'])

        selected_function, required_length = cli_moving_averages, 1

    # FUNCTION: Optimize Portfolio Variance/Volatility
    elif args['function_arg'] in definitions.FUNC_DICT['optimize_portfolio']['values']:
        def cli_optimize_portfolio_variance():
            from scrilla.analysis.objects.portfolio import Portfolio

            portfolio = Portfolio(tickers=args['tickers'],
                                  start_date=args['start_date'],
                                  end_date=args['end_date'],
                                  method=args['estimation_method'])

            if args['optimize_sharpe']:
                from scrilla.analysis.optimizer import maximize_sharpe_ratio
                allocation = maximize_sharpe_ratio(
                    portfolio=portfolio, target_return=args['target'])
            else:
                from scrilla.analysis.optimizer import optimize_portfolio_variance
                allocation = optimize_portfolio_variance(
                    portfolio=portfolio, target_return=args['target'])

            if args['investment'] is not None:
                from scrilla.services import get_daily_prices_latest
                prices = get_daily_prices_latest(
                    tickers=args['tickers'])
            else:
                prices = None

            if print_format_to_screen(args):
                from scrilla.util.outputter import optimal_result
                optimal_result(
                    portfolio=portfolio, allocation=allocation, investment=args['investment'])

            if print_json_to_screen(args):
                from scrilla.static.formats import format_allocation
                from json import dumps
                print(dumps(format_allocation(allocation=allocation,
                                              portfolio=portfolio,
                                              investment=args['investment'],
                                              latest_prices=prices)))

            if args['save_file'] is not None:
                from scrilla.files import save_allocation
                save_allocation(allocation=allocation,
                                portfolio=portfolio,
                                file_name=args['save_file'],
                                investment=args['investment'],
                                latest_prices=prices)

        selected_function, required_length = cli_optimize_portfolio_variance, 2

    # FUNCTION: Optimize Portfolio Conditional Value At Risk
    elif args['function_arg'] in definitions.FUNC_DICT['optimize_portfolio_conditional_var']['values']:
        def cli_optimize_conditional_value_at_risk():
            from scrilla.analysis.optimizer import optimize_conditional_value_at_risk
            from scrilla.analysis.objects.portfolio import Portfolio

            portfolio = Portfolio(tickers=args['tickers'],
                                  start_date=args['start_date'],
                                  end_date=args['end_date'],
                                  method=args['estimation_method'])
            allocation = optimize_conditional_value_at_risk(portfolio=portfolio,
                                                            prob=args['probability'],
                                                            expiry=args['expiry'],
                                                            target_return=args['target'])
            if args['investment'] is not None:
                from scrilla.services import get_daily_prices_latest
                prices = get_daily_prices_latest(
                    tickers=args['tickers'])
            else:
                prices = None

            if print_format_to_screen(args):
                from scrilla.util.outputter import optimal_result
                optimal_result(portfolio=portfolio,
                               allocation=allocation,
                               investment=args['investment'],
                               latest_prices=prices)

            if print_json_to_screen(args):
                from scrilla.static.formats import format_allocation
                from json import dumps
                print(dumps(format_allocation(allocation=allocation,
                                              portfolio=portfolio,
                                              investment=args['investment'],
                                              latest_prices=prices)))

            if args['save_file'] is not None:
                from scrilla.files import save_allocation
                save_allocation(allocation=allocation,
                                portfolio=portfolio,
                                file_name=args['save_file'],
                                investment=args['investment'],
                                latest_prices=prices)
        selected_function, required_length = cli_optimize_conditional_value_at_risk, 2

    # FUNCTION: Plot Correlation Time Series
    elif args['function_arg'] in definitions.FUNC_DICT['plot_correlation']['values']:
        def cli_plot_correlation():
            from scrilla.analysis.models.geometric.statistics import calculate_moment_correlation_series
            from scrilla.analysis.plotter import plot_correlation_series

            logger.comment('This calculation takes a while, strap in...')

            correlation_history = calculate_moment_correlation_series(ticker_1=args['tickers'][0],
                                                                      ticker_2=args['tickers'][1],
                                                                      start_date=args['start_date'],
                                                                      end_date=args['end_date'])
            plot_correlation_series(tickers=args['tickers'],
                                    series=correlation_history,
                                    savefile=args['save_file'])

        selected_function, required_length, exact = cli_plot_correlation, 2, True

    # FUNCTION: Plot Dividend History With Linear Regression Model
    elif args['function_arg'] in definitions.FUNC_DICT['plot_dividends']['values']:
        def cli_plot_dividends():
            from scrilla.services import get_dividend_history
            from scrilla.analysis.objects.cashflow import Cashflow
            from scrilla.analysis.plotter import plot_cashflow

            dividends = get_dividend_history(ticker=args['tickers'][0])
            if args['discount'] is None:
                from scrilla.analysis.markets import cost_of_equity
                args['discount'] = cost_of_equity(ticker=args['tickers'][0],
                                                  method=args['estimation_method'])
            div_cashflow = Cashflow(sample=dividends,
                                    discount_rate=args['discount'])
            plot_cashflow(ticker=args['tickers'][0],
                          cashflow=div_cashflow, show=True,
                          savefile=args['save_file'])

        selected_function, required_length, exact = cli_plot_dividends, 1, True

    # FUNCTION: Plot Efficient Frontier
    elif args['function_arg'] in definitions.FUNC_DICT['plot_frontier']['values']:
        def cli_plot_frontier():
            from scrilla.analysis.objects.portfolio import Portfolio
            from scrilla.analysis.optimizer import calculate_efficient_frontier
            from scrilla.analysis.plotter import plot_frontier

            portfolio = Portfolio(tickers=args['tickers'],
                                  start_date=args['start_date'],
                                  end_date=args['end_date'],
                                  method=args['estimation_method'])

            frontier = calculate_efficient_frontier(portfolio=portfolio,
                                                    steps=args['steps'])

            plot_frontier(portfolio=portfolio,
                          frontier=frontier,
                          show=True,
                          savefile=args['save_file'])

        selected_function, required_length = cli_plot_frontier, 2

    # FUNCTION: Plot Moving Averages of Logarithmic Returns
    elif args['function_arg'] in definitions.FUNC_DICT['plot_moving_averages']['values']:
        def cli_plot_moving_averages():
            from scrilla.analysis.models.geometric.statistics import calculate_moving_averages
            from scrilla.analysis.plotter import plot_moving_averages
            # TODO: estimation techniques with moving averages
            moving_averages = calculate_moving_averages(ticker=args['tickers'][0],
                                                        start_date=args['start_date'],
                                                        end_date=args['end_date'],
                                                        method=args['estimation_method'])

            plot_moving_averages(ticker=args['tickers'][0],
                                 averages=moving_averages,
                                 show=True, savefile=args['save_file'])

        selected_function, required_length, exact = cli_plot_moving_averages, 1, True

    # FUNCTION: Plot Return QQ Series
    elif args['function_arg'] in definitions.FUNC_DICT['plot_return_qq']['values']:
        def cli_plot_qq_returns():
            from scrilla.analysis.models.geometric.statistics import get_sample_of_returns
            from scrilla.analysis.estimators import qq_series_for_sample
            from scrilla.analysis.plotter import plot_qq_series

            returns = get_sample_of_returns(ticker=args['tickers'][0],
                                            start_date=args['start_date'],
                                            end_date=args['end_date'],
                                            daily=True)

            qq_series = qq_series_for_sample(sample=returns)

            plot_qq_series(ticker=args['tickers'][0],
                           qq_series=qq_series,
                           show=True,
                           savefile=args['save_file'])

        selected_function, required_length, exact = cli_plot_qq_returns, 1, True

    # FUNCTION: Plot Histogram of Returns
    elif args['function_arg'] in definitions.FUNC_DICT['plot_return_dist']['values']:
        def cli_plot_dist_returns():
            from scrilla.analysis.models.geometric.statistics import get_sample_of_returns
            from scrilla.analysis.plotter import plot_return_histogram

            returns = get_sample_of_returns(ticker=args['tickers'][0],
                                            start_date=args['start_date'],
                                            end_date=args['end_date'],
                                            daily=True)
            plot_return_histogram(ticker=args['tickers'][0],
                                  sample=returns,
                                  show=True,
                                  savefile=args['save_file'])
        selected_function, required_length, exact = cli_plot_dist_returns, 1, True

    # FUNCTION: Plot Risk-Return Profile
    elif args['function_arg'] in definitions.FUNC_DICT['plot_risk_profile']['values']:
        def cli_plot_risk_profile():
            from scrilla.analysis.models.geometric.statistics import calculate_risk_return
            from scrilla.analysis.plotter import plot_profiles
            from scrilla.util.dater import format_date_range
            profiles = {}
            for arg in args['tickers']:
                profiles[arg] = calculate_risk_return(ticker=arg,
                                                      start_date=args['start_date'],
                                                      end_date=args['end_date'],
                                                      method=args['estimation_method'])

            plot_profiles(symbols=args['tickers'],
                          show=True,
                          profiles=profiles,
                          savefile=args['save_file'],
                          subtitle=format_date_range(start_date=args['start_date'],
                                                     end_date=args['end_date']))
        selected_function, required_length = cli_plot_risk_profile, 1

    elif args['function_arg'] in definitions.FUNC_DICT['plot_yield_curve']['values']:
        def cli_plot_yield_curve():
            from scrilla.util.dater import get_next_business_date, to_string
            from scrilla.static.keys import keys
            from scrilla.services import get_daily_interest_history
            from scrilla.analysis.plotter import plot_yield_curve
            yield_curve = {}
            args['start_date'] = get_next_business_date(args['start_date'])
            start_date_string = to_string(args['start_date'])
            yield_curve[start_date_string] = []
            for maturity in keys['YIELD_CURVE']:
                rate = get_daily_interest_history(maturity=maturity,
                                                  start_date=args['start_date'],
                                                  end_date=args['start_date'])
                yield_curve[start_date_string].append(rate[start_date_string])

            plot_yield_curve(yield_curve=yield_curve,
                             show=True,
                             savefile=args['save_file'])
        selected_function, required_length, exact = cli_plot_yield_curve, 0, True
    # FUNCTION: Price History
    elif args['function_arg'] in definitions.FUNC_DICT['price_history']['values']:
        def cli_price_history():
            from scrilla.services import get_daily_price_history
            from scrilla.static.keys import keys
            if print_format_to_screen(args):
                from scrilla.util.outputter import scalar_result

            all_prices = {}
            for arg in args['tickers']:
                prices = get_daily_price_history(ticker=arg,
                                                 start_date=args['start_date'],
                                                 end_date=args['end_date'])
                all_prices[arg] = {}
                for this_date in prices:
                    price = prices[this_date][keys['PRICES']['CLOSE']]
                    all_prices[arg][this_date] = price

                    if print_format_to_screen(args):
                        scalar_result(
                            calculation=f'{arg}({this_date})', result=float(price))

            if print_json_to_screen(args):
                from json import dumps
                print(dumps(all_prices))

            if args['save_file'] is not None:
                from scrilla.files import save_file
                save_file(file_to_save=all_prices, file_name=args['save_file'])
        selected_function, required_length = cli_price_history, 1

    # FUNCTION: Interest Rate History
    elif args['function_arg'] in definitions.FUNC_DICT['interest_history']['values']:
        def cli_interest_history():
            from scrilla.services import get_daily_interest_history
            if print_format_to_screen(args):
                from scrilla.util.outputter import scalar_result

            all_rates = {}
            for arg in args['tickers']:
                all_rates[arg] = get_daily_interest_history(maturity=arg,
                                                            start_date=args['start_date'],
                                                            end_date=args['end_date'])
                for this_date in all_rates[arg]:
                    all_rates[arg][this_date] = all_rates[arg][this_date]

                if print_format_to_screen(args):
                    for this_date in all_rates[arg]:
                        scalar_result(calculation=f'{arg}_YIELD({this_date})', result=float(
                            all_rates[arg][this_date])/100, currency=False)

            if print_json_to_screen(args):
                from json import dumps
                print(dumps(all_rates))

            if args['save_file'] is not None:
                from scrilla.files import save_file
                save_file(file_to_save=all_rates,
                          file_name=args['save_file'])
        selected_function, required_length = cli_interest_history, 1

    # FUNCTION: Risk Free Rate
    elif args['function_arg'] in definitions.FUNC_DICT['risk_free_rate']['values']:
        def cli_risk_free_rate():
            from scrilla.services import get_risk_free_rate
            from scrilla.settings import RISK_FREE_RATE
            rate = {}
            rate[RISK_FREE_RATE] = get_risk_free_rate()

            if print_format_to_screen(args):
                from scrilla.util.outputter import title_line, scalar_result
                from scrilla.static.formats import formats
                title_line("Risk Free Rate")
                scalar_result(calculation=formats['RISK_FREE_TITLE'].format(RISK_FREE_RATE),
                              result=rate[RISK_FREE_RATE], currency=False)
            if print_json_to_screen(args):
                from json import dumps
                print(dumps(rate))

            if args['save_file'] is not None:
                from scrilla.files import save_file
                save_file(file_to_save=rate, file_name=args['save_file'])
        selected_function, required_length, exact = cli_risk_free_rate, 0, True

    # FUNCTION: Risk-Return Profile
    elif args['function_arg'] in definitions.FUNC_DICT["risk_profile"]['values']:
        def cli_risk_return():
            from scrilla.analysis.models.geometric.statistics import calculate_risk_return
            from scrilla.analysis.markets import sharpe_ratio, market_beta, cost_of_equity
            profiles = {}
            for arg in args['tickers']:
                profiles[arg] = calculate_risk_return(ticker=arg,
                                                      method=args['estimation_method'],
                                                      start_date=args['start_date'],
                                                      end_date=args['end_date'])
                profiles[arg]['sharpe_ratio'] = sharpe_ratio(ticker=arg,
                                                             start_date=args['start_date'],
                                                             end_date=args['end_date'],
                                                             ticker_profile=profiles[arg],
                                                             method=args['estimation_method'])
                profiles[arg]['asset_beta'] = market_beta(ticker=arg,
                                                          start_date=args['start_date'],
                                                          end_date=args['end_date'],
                                                          ticker_profile=profiles[arg],
                                                          method=args['estimation_method'])
                profiles[arg]['equity_cost'] = cost_of_equity(ticker=arg,
                                                              start_date=args['start_date'],
                                                              end_date=args['end_date'],
                                                              method=args['estimation_method'])

            if print_format_to_screen(args):
                from scrilla.util.outputter import risk_profile
                risk_profile(profiles=profiles)

            if print_json_to_screen(args):
                from json import dumps
                print(dumps(profiles))

            if args['save_file'] is not None:
                from scrilla.files import save_file
                save_file(file_to_save=profiles, file_name=args['save_file'])

        selected_function, required_length = cli_risk_return, 1

    # FUNCTION: Model Discount Screener
    elif args['function_arg'] in definitions.FUNC_DICT["screener"]['values']:
        def cli_screener():
            from scrilla.analysis.markets import screen_for_discount
            from scrilla.static.keys import keys
            from scrilla.util.outputter import screen_results
            if args['model'] is None:
                model = keys['MODELS']['DDM']
            results = screen_for_discount(
                model=model, discount_rate=args['discount'])
            screen_results(info=results, model=model)
        selected_function, required_length = cli_screener, 0

    # FUNCTION: Sharpe Ratio
    elif args['function_arg'] in definitions.FUNC_DICT["sharpe_ratio"]['values']:
        def cli_sharpe_ratio():
            from scrilla.analysis.markets import sharpe_ratio
            all_results = {}
            for arg in args['tickers']:
                result = sharpe_ratio(ticker=arg,
                                      start_date=args['start_date'],
                                      end_date=args['end_date'],
                                      method=args['estimation_method'])
                all_results[arg] = result

                if print_format_to_screen(args):
                    from scrilla.util.outputter import scalar_result
                    scalar_result(calculation=f'{arg}_sharpe_ratio', result=result,
                                  currency=False)

            if print_json_to_screen(args):
                from json import dumps
                print(dumps(all_results))

            if args['save_file'] is not None:
                from scrilla.files import save_file
                save_file(file_to_save=all_results,
                          file_name=args['save_file'])
        selected_function, required_length = cli_sharpe_ratio, 1

   # FUNCTION: Store Key
    elif args['function_arg'] in definitions.FUNC_DICT['store']['values']:
        def cli_store():
            from scrilla.files import set_credentials
            set_credentials(value=args['value'], which_key=args['key'])
        selected_function, required_length = cli_store, 0

    # FUNCTION: Get Latest Economic Statistic
    elif args['function_arg'] in definitions.FUNC_DICT["statistic"]['values']:
        def cli_statistic():
            from scrilla.services import get_daily_fred_latest
            all_stats = {}
            for stat in args['tickers']:
                result = get_daily_fred_latest(symbol=stat)
                all_stats[stat] = result

                if print_format_to_screen(args):
                    from scrilla.util.outputter import scalar_result
                    scalar_result(calculation=stat,
                                  result=result, currency=False)

            if print_json_to_screen(args):
                from json import dumps
                print(dumps(all_stats))

            if args['save_file'] is not None:
                from scrilla.files import save_file
                save_file(file_to_save=all_stats, file_name=args['save_file'])

        selected_function, required_length = cli_statistic, 1

    # FUNCTION: Statistic History
    elif args['function_arg'] in definitions.FUNC_DICT['statistic_history']['values']:
        def cli_statistic_history():
            from scrilla.services import get_daily_fred_history
            if print_format_to_screen(args):
                from scrilla.util.outputter import scalar_result

            all_stats = {}
            for arg in args['tickers']:
                stats = get_daily_fred_history(symbol=arg,
                                               start_date=args['start_date'],
                                               end_date=args['end_date'])
                all_stats[arg] = stats
                if print_format_to_screen(args):
                    for this_date in stats:
                        scalar_result(calculation=f'{arg}({this_date})',
                                      result=stats[this_date],
                                      currency=False)

            if print_json_to_screen(args):
                from json import dumps
                print(dumps(all_stats))

            if args['save_file'] is not None:
                from scrilla.files import save_file
                save_file(file_to_save=all_stats, file_name=args['save_file'])
        selected_function, required_length = cli_statistic_history, 1

    # FUNCTION: Set Watchlist
    elif args['function_arg'] in definitions.FUNC_DICT["watchlist"]['values']:
        def cli_watchlist():
            from scrilla.files import add_watchlist
            add_watchlist(new_tickers=args['tickers'])
            logger.comment(
                "Watchlist saved. Use -ls option to print watchlist.")
        selected_function, required_length = cli_watchlist, 1

    else:
        def cli_help():
            from scrilla.util.outputter import help_msg
            help_msg()
        selected_function, required_length = cli_help, 0

    # END CLI FUNCTION DEFINITIONS

    if selected_function is not None:
        validate_function_usage(selection=args['function_arg'],
                                args=args['tickers'],
                                wrapper_function=selected_function,
                                required_length=required_length,
                                exact=exact)


def scrilla():
    import sys

    do_program(sys.argv[1:])


if __name__ == "__main__":
    import sys

    do_program(sys.argv[1:])



# This file is part of scrilla: https://github.com/chinchalinchin/scrilla.

# scrilla is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3
# as published by the Free Software Foundation.

# scrilla is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with scrilla.  If not, see <https://www.gnu.org/licenses/>
# or <https://github.com/chinchalinchin/scrilla/blob/develop/main/LICENSE>.

import sys

from PySide6 import QtCore, QtWidgets, QtGui

from scrilla import settings

from scrilla.gui import formats, utilities, definitions
from scrilla.gui.widgets import factories

# NOTE: widget_buttons and function_widgets must preserve order.


class MenuWidget(QtWidgets.QWidget):
    """

    .. notes::
        * Widget Hierarchy: 

    .. links::
        * [Stack Overflow: Weird Lambda Behavior in Loops](https://stackoverflow.com/questions/19768456/weird-lambda-behaviour-in-loops)
    """

    def __init__(self):
        super().__init__()
        self._init_menu_widgets()
        self._generate_menu_bar()
        self._generate_splash()
        self._arrange_menu_widgets()
        self._stage_menu_widgets()

    def _generate_menu_bar(self):
        self.menu_bar = factories.atomic_widget_factory(
            component='menu-bar', title=None)
        self.menus = []

        for j, menu in enumerate(definitions.MENUBAR_WIDGET):
            self.menus.append(self.menu_bar.addMenu(menu))
            for i, action in enumerate(definitions.MENUBAR_WIDGET[menu]):
                q_action = QtGui.QAction(action['name'], self)
                q_action.setShortcut(action['shortcut'])
                if menu == 'Functions':
                    q_action.triggered.connect(
                        (lambda i: lambda: self._show_widget(i))(i))
                elif menu == 'Account':
                    pass
                elif menu == 'View':
                    if action['name'] == 'Function Menu':
                        q_action.triggered.connect(lambda: self.function_menu.setVisible(
                            (not self.function_menu.isVisible())))
                    elif action['name'] == 'Splash Menu':
                        q_action.triggered.connect(self._clear)

                self.menus[j].addAction(q_action)

    def _generate_splash(self):
        self.splash_container = factories.layout_factory(layout='vertical-box')
        self.source_button = factories.atomic_widget_factory(
            component='source-button', title=None)
        self.package_button = factories.atomic_widget_factory(
            component='package-button', title=None)
        self.documentation_button = factories.atomic_widget_factory(
            component='documentation-button', title=None)
        self.splash = factories.atomic_widget_factory(
            component='splash', title=None)

        self.splash_button_panel = factories.layout_factory(
            layout='horizontal-box')

        self.splash_button_panel.layout().addStretch()
        self.splash_button_panel.layout().addWidget(self.source_button)
        self.splash_button_panel.layout().addWidget(self.package_button)
        self.splash_button_panel.layout().addWidget(self.documentation_button)
        self.splash_button_panel.layout().addStretch()

    def _init_menu_widgets(self):
        self.setObjectName('root')
        self.title = factories.atomic_widget_factory(
            component='title', title=settings.APP_NAME)

        self.function_menu = factories.layout_factory(layout='vertical-box')
        self.function_menu.setObjectName('grand-child')
        self.function_title_container = factories.layout_factory(
            layout='horizontal-box')
        self.function_title = factories.atomic_widget_factory(
            component='heading', title='Functions')
        self.function_hide_button = factories.atomic_widget_factory(
            component='hide-button', title=None)

        self.widget_buttons = [factories.atomic_widget_factory(
            component='button', title=function['name']) for function in definitions.FUNC_WIDGETS.values()]
        self.exit_button = factories.atomic_widget_factory(
            component='button', title="Exit")

        self.function_widgets = [function['class'](
            'great-grand-child', self) for function in definitions.FUNC_WIDGETS.values()]

        self.display_pane = factories.layout_factory(layout='vertical-box')
        self.display_pane.setObjectName('grand-child')

        self.container_pane = factories.layout_factory(layout='horizontal-box')
        self.container_pane.setObjectName('child')

        self.setLayout(QtWidgets.QVBoxLayout())

    def _arrange_menu_widgets(self):
        """Arranges child widget within their layouts."""
        self.title.setAlignment(QtCore.Qt.AlignHCenter)
        # self.splash.setAlignment(QtCore.Qt.AlignCenter)

        self.container_pane.setSizePolicy(QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding,
                                                                QtWidgets.QSizePolicy.Expanding))
        self.function_menu.setSizePolicy(QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum,
                                                               QtWidgets.QSizePolicy.Expanding))
        self.display_pane.setSizePolicy(QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding,
                                                              QtWidgets.QSizePolicy.Expanding))
        self.title.setSizePolicy(QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding,
                                                       QtWidgets.QSizePolicy.Minimum))

        self.function_title_container.layout().addWidget(self.function_title)
        self.function_title_container.layout().addWidget(self.function_hide_button)

        self.function_menu.layout().addWidget(self.function_title_container)
        for button in self.widget_buttons:
            self.function_menu.layout().addWidget(button)
        self.function_menu.layout().addStretch()
        self.function_menu.layout().addWidget(self.exit_button)

        self.splash_container.layout().addWidget(self.splash)
        self.splash_container.layout().addStretch()
        self.splash_container.layout().addWidget(self.splash_button_panel)

        self.display_pane.layout().addWidget(self.splash_container)
        for widget in self.function_widgets:
            self.display_pane.layout().addWidget(widget)

        self.container_pane.layout().addWidget(self.function_menu)
        self.container_pane.layout().addWidget(self.display_pane)

        self.layout().addWidget(self.menu_bar)
        self.layout().addWidget(self.title)
        self.layout().addWidget(self.container_pane)

    def _stage_menu_widgets(self):
        for i, button in enumerate(self.widget_buttons):
            button.show()
            # see #NOTE
            button.clicked.connect((lambda i: lambda: self._show_widget(i))(i))

        for widget in self.function_widgets:
            widget.hide()

        self.function_hide_button.clicked.connect(
            lambda: utilities.switch_visibility(self.function_menu))
        self.exit_button.clicked.connect(self.close)
        self.source_button.clicked.connect(
            lambda: utilities.open_browser(utilities.get_metadata('source')))
        self.package_button.clicked.connect(
            lambda: utilities.open_browser(utilities.get_metadata('package')))
        self.documentation_button.clicked.connect(
            lambda: utilities.open_browser(utilities.get_metadata('documentation')))
        self.exit_button.show()
        self.menu_bar.show()

    def _clear(self):
        for widget in self.function_widgets:
            widget.hide()
        self.splash_container.show()

    @QtCore.Slot()
    def _show_widget(self, widget_index):
        self.splash_container.hide()
        for widget in self.function_widgets:
            widget.hide()
        self.function_widgets[widget_index].show()
        self.title.setText(list(definitions.FUNC_WIDGETS.values())[
                           widget_index]['name'])


if __name__ == "__main__":
    app = QtWidgets.QApplication([])

    mainWidget = MenuWidget()
    mainWidget.resize(settings.GUI_WIDTH, settings.GUI_HEIGHT)
    mainWidget.show()

    sys.exit(app.exec_())



constants = {
    'ONE_TRADING_DAY': {
        'EQUITY': (1/252),
        'CRYPTO': (1/365)
    },
    'ACCURACY': 7,
    'BACKOFF_PERIOD': 30,
    'KEEP_FILE': '.gitkeep',
    'PRICE_YEAR_CUTOFF': 1950,
    'DENOMINATION': 'USD',
    'NPV_DELTA_TOLERANCE': 0.0000001,
    'OPTIMIZATION_METHOD': "SLSQP",
    'SIG_FIGS': 5,
    'LOG_LEVEL': {
        'NONE': 'none',
        'INFO': 'info',
        'DEBUG': 'debug',
        'VERBOSE': 'verbose'
    }
}



# This file is part of scrilla: https://github.com/chinchalinchin/scrilla.

# scrilla is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3
# as published by the Free Software Foundation.

# scrilla is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with scrilla.  If not, see <https://www.gnu.org/licenses/>
# or <https://github.com/chinchalinchin/scrilla/blob/develop/main/LICENSE>.
"""
A series of classes that all inherit from ``PySide6.QtWidgets.QWidget` and build up in sequence the necessary functionality to grab and validate user input, calculate and display results, etc. These widgets are not directly displayed on the GUI; rather, they are used as building blocks by the `scrilla.gui.functions` module, to create widgets specifically for application functions.

All widgets have a similar structure in that they call a series of methods tailored to their specific class: 
    1. **_init**
        Child widgets are constructed. Layouts are created.
    2. **_arrange**
        Child widgets are arranged. Layouts are applied.
    3. **_stage**
        Child widgets are prepped for display.

The idea behind the module is to hide as much widget initialization and styling as possible behind its functions, so the `scrilla.gui.functions` module can focus on application-level logic and isn't cluttered with ugly declarative configuration.

Widgets have styles applied to them through the `scrilla.styles.app.qss` stylesheet. Widgets are layered over one another in a hierarchy described by: `root` -> `child` -> `grand-child` -> `great-grand-child` -> etc. Each layers has a theme that descends down the material color scheme hex codes in sequence with its place in the layer hierarchy. See `scrilla.gui.formats` for more information. 
"""
import datetime
from typing import Callable, Dict, List, Union
from PySide6 import QtCore, QtWidgets, QtGui

from scrilla import settings
from scrilla.util import helper
from scrilla.gui import utilities
from scrilla.static import definitions
from scrilla.gui.widgets import factories

SYMBOLS_LIST = "list"
"""Constant passed into `scrilla.gui.widgets.components.ArgumentWidget` to initialize an input control allowing user to input list of ticker symbols"""
SYMBOLS_SINGLE = "single"
"""Constant passed into `scrilla.gui.widgets.components.ArgumentWidget` to initialize an input control allowing user to input  single ticker symbol"""
SYMBOLS_NONE = "none"
"""Constant passed into `scrilla.gui.widgets.components.ArgumentWidget` to initialize an input control without ticker symbols"""


class SkeletonWidget(QtWidgets.QWidget):
    def __init__(self, function: str, parent: QtWidgets.QWidget):
        super(SkeletonWidget, self).__init__(parent)
        self._configure_control_skeleton(function)

    def _configure_control_skeleton(self, function: str):
        self.controls = factories.generate_control_skeleton()

        for arg in definitions.FUNC_DICT[function]['args']:
            if not definitions.ARG_DICT[arg]['cli_only']:
                self.controls[arg] = True


class ArgumentWidget(QtWidgets.QWidget):
    """
    Base class for other more complex widgets to inherit. An instance of `scrilla.gui.widgets.ArgumentWidget` embeds its child widgets in its own layout,  which is an instance of``PySide6.QtWidgetQVBoxLayout``. This class provides access to the following widgets:a ``PySide6.QtWidgets.QLabel`` for the title widget, an error message widget, a ``PySide6.QtWidgets.QPushButton`` for the calculate button widget, a ``PySide6.QtWidget.QLineEdit`` for ticker symbol input, and a variety of optional input widgets enabled by boolean flags in the `controls` constructor argument. 

    Constructor
    -----------
    1. **calculate_function**: ``str``
        Function attached to the `calculate_button` widget. Triggered when the widget is clicked.
    2. **clear_function**: ``str``
        Function attached to the `clear_button` widget. Trigged when the widget is clicked.
    3. **controls:**: ``Dict[bool]``
        Dictionary of boolean flags instructing the class which optional input to include. Optional keys can be found in `scrilla.static.definitions.ARG_DICT`. An dictionary skeleton with all the optional input disabled can be retrieved through `scrilla.gui.widgets.factories.generate_control_skeleton`.
    4. **layer**: ``str``
        Stylesheet property attached to widget.
    5. **single**: ``bool``
        Flag to limit symbol input to one ticker symbol. Calls to `get_symbol_input()` will return 

    Attributes
    ----------
    1. **title**: ``PySide6.QtWidgets.QLabel``
    2. **required_title**: ``PySide6.QtWidgets.QLabel``
    3. **optional_title**: ``PySide6.QtWidgets.QLabel``
    4. **required_pane**: ``PySide6.QtWidgets.QLabel``
    5. **optional_pane**: ``PySide6.QtWidgets.QLabel``
    6. **message**: ``PySide6.QtWidgets.QLabel``
        Label attached to `self.symbol_input`.
    7. **error_message**: ``PySide6.QtWidget.QLabel``
        Message displayed if there is an error thrown during calculation.
    8. **calculate_button**: ``PySide6.QtWidget.QPushButton``
        Empty button for super classes to inherit, primed to fire events on return presses
    9. **clear_button**: ``PySide6.QtWidget.QPushButton``
        Empty button for super classes to inherit, primed to fire events on return presses.
    10. **symbol_widget**: ``PySide6.QtWidget.QLineEdit``
        Empty text area for super classes to inherit
    11. **control_widgets**: ``Dict[str,Union[PySide6.QtWidgets.QWidget,None]]
        Dictionary containing the optional input widgets.
    """
    # TODO: calculate and clear should be part of THIS constructor, doesn't make sense to have other widgets hook them up.

    def __init__(self, calculate_function: Callable, clear_function: Callable, controls: Dict[str, bool], layer, mode: str = SYMBOLS_LIST):
        super().__init__()
        self.layer = layer
        self.controls = controls
        self.control_widgets = {}
        self.group_control_widgets = None
        self.calculate_function = calculate_function
        self.clear_function = clear_function

        self._init_widgets(mode)
        self._generate_group_widgets()
        self._arrange_widgets()
        self._stage_widgets()

    def _init_widgets(self, mode: str) -> None:
        """
        Creates child widgets by calling factory methods from `scrilla.gui.widgets.factories`. This method will iterate over `self.controls` and initialize the optional input widget accordingly. `self.optional_pane` and `self.required_pane`, the container widgets for the input elements, are initialized with a style tag on the same layer as the `scrilla.gui.widgets.components.ArgumentWidget`.
        """
        self.title = factories.atomic_widget_factory(
            component='subtitle', title='Function Input')
        self.optional_title = factories.atomic_widget_factory(
            component='label', title='Optional Arguments')
        self.error_message = factories.atomic_widget_factory(
            component='error', title="Error Message Goes Here")
        self.calculate_button = factories.atomic_widget_factory(
            component='calculate-button', title='Calculate')
        self.clear_button = factories.atomic_widget_factory(
            component='clear-button', title='Clear')

        if mode == SYMBOLS_LIST:
            self.required_title = factories.atomic_widget_factory(
                component='label', title='Required Arguments')
            self.symbol_hint = factories.atomic_widget_factory(
                component='text', title="Separate Tickers With Commas")
            self.required_pane = factories.layout_factory(
                layout='vertical-box')
            self.required_pane.setObjectName(self.layer)
            self.symbol_widget = factories.argument_widget_factory(
                component='symbols', title="Asset Tickers :", optional=False)
        elif mode == SYMBOLS_SINGLE:
            self.required_title = factories.atomic_widget_factory(
                component='label', title='Required Argument')
            self.symbol_hint = factories.atomic_widget_factory(
                component='text', title="Enter a Single Symbol")
            self.required_pane = factories.layout_factory(
                layout='vertical-box')
            self.required_pane.setObjectName(self.layer)
            self.symbol_widget = factories.argument_widget_factory(
                component='symbol', title="Symbol: ", optional=False)
        else:
            self.symbol_widget = None

        self.group_definitions = None
        for control in self.controls:
            if self.controls[control]:
                if self.controls[control]:
                    if definitions.ARG_DICT[control]["widget_type"] != "group":
                        self.control_widgets[control] = factories.argument_widget_factory(definitions.ARG_DICT[control]['widget_type'],
                                                                                          f'{definitions.ARG_DICT[control]["name"]} :',
                                                                                          optional=True)
                    else:
                        if self.group_definitions is None:
                            self.group_definitions = {}
                        self.group_definitions[definitions.ARG_DICT[control]
                                               ['name']] = definitions.ARG_DICT[control]
            else:
                self.control_widgets[control] = None

        self.optional_pane = factories.layout_factory(layout='vertical-box')
        self.optional_pane.setObjectName(self.layer)
        self.setLayout(QtWidgets.QVBoxLayout())

    def _generate_group_widgets(self):
        if self.group_definitions is not None:
            self.group_control_widgets = {}
            groups = {}
            for definition in self.group_definitions:
                group_key = self.group_definitions[definition]['group']
                group_name = definitions.GROUP_DICT[group_key]

                if group_name not in groups.keys():
                    groups[group_name] = [definition]
                else:
                    groups[group_name].append(definition)

            for group_name, group in groups.items():
                self.group_control_widgets[group_name] = factories.group_widget_factory(
                    group, group_name)

    def _arrange_widgets(self):
        """
        Arrange child widgets in their layouts and provides rendering hints. The `self.symbol_widget` is set into a ``PySide6.QtWidgets.QVBoxLayout`` named `self.required_pane`. The optional input widgets are set into a separate ``PySide6.QtWidgets.QVBoxLayout`` named `self.optional_pane`. `self.required_pane` and `self.optional_pane` are in turn set into a parent `PySide6.QtWidgets.QVBoxLayout``, along with `self.calculate_button` and `self.clear_button`. A strecth widget is inserted between the input widgets and the button widgets. 

        """
        factories.set_policy_on_widget_list([self.title, self.optional_title, self.optional_pane],
                                            QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Maximum))
        if self.symbol_widget is not None:
            factories.set_policy_on_widget_list([self.symbol_hint, self.required_title, self.required_pane],
                                                QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Maximum))
        self.setSizePolicy(QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding))

        if self.symbol_widget is not None:
            self.required_pane.layout().addWidget(self.required_title)
            self.required_pane.layout().addWidget(self.symbol_hint)
            self.required_pane.layout().addWidget(self.symbol_widget)
        if self.group_control_widgets is not None:
            for widget in self.group_control_widgets.values():
                self.required_pane.layout().addWidget(widget)

        self.optional_pane.layout().addWidget(self.optional_title)
        for control_widget in self.control_widgets.values():
            if control_widget is not None:
                self.optional_pane.layout().addWidget(control_widget)

        self.layout().addWidget(self.title)
        self.layout().addWidget(self.error_message)
        if self.symbol_widget is not None:
            self.layout().addWidget(self.required_pane)
        self.layout().addWidget(self.optional_pane)
        self.layout().addStretch()
        self.layout().addWidget(self.calculate_button)
        self.layout().addWidget(self.clear_button)

    def _stage_widgets(self):
        """
        Prepares child widgets for display. `self.clear_function` and `self.calculate` function are hooked into the `clicked` signal emitted from `self.clear_button` and `self.calculate_button`, respectively.
        """
        self.error_message.hide()
        self.clear_button.clicked.connect(self.clear_function)
        self.calculate_button.clicked.connect(self.calculate_function)
        if self.symbol_widget is not None:
            # NOTE: symbol widget is technically a layout in which the lineedit is abutted by a label, so need
            # to pull the actual input element from the layout
            self.symbol_widget.layout().itemAt(1).widget(
            ).returnPressed.connect(self.calculate_function)

    def get_symbol_input(self) -> Union[List[str], None]:
        """
        Returns the symbols inputted into the `PySide6.QtWidgets.QLineEdit` child of `symbol_widget`. If the `ArgumentWidget` has been initialized as a `scrilla.gui.widgets.components.SYMBOLS_LIST`, the method will return a list of inputted strings. If the `ArgumentWidget` has been initialied as a `scrilla.gui.widgets.components.SYMBOLS_SINGLE`, the method will return a list with a string as its single element. If the `ArgumentWidget` has been initialized as a `scrilla.gui.widgets.components.SYMBOLS_NONE`, this method will return `None`. In addition, in the first two cases, if no input has been entered, this method will return `None`. 
        """
        if self.symbol_widget is not None:
            return helper.split_and_strip(self.symbol_widget.layout().itemAt(1).widget().text())
        return None

    def get_control_input(self, control_widget_key: str) -> Union[datetime.date, str, bool, None]:
        """
        Get the value on the specified optional input widget. Optional keys are accessed through the keys of the `scrilla.static.definitions.ARG_DICT` dictionary.

        If the widget is disabled or has been excluded altogether from the parent widget, i.e. a value of `False` was passed in through the constructor's `controls` arguments for that optional input widget, this method will return `None`.
        """
        if self.control_widgets[control_widget_key] is None:
            return None

        widget = self.control_widgets[control_widget_key].layout().itemAt(
            1).widget()

        if not widget.isEnabled():
            return None

        if type(widget) is QtWidgets.QDateEdit:
            return widget.date().toPython()
        if type(widget) is QtWidgets.QLineEdit:
            if type(widget.validator()) is QtGui.QIntValidator:
                return int(widget.text())
            if type(widget.validator()) in [QtGui.QDoubleValidator, QtGui.QRegularExpressionValidator]:
                return float(widget.text())
            return widget.text()
        if type(widget) is QtWidgets.QRadioButton:
            return widget.isChecked()

    def prime(self) -> None:
        """
        Enables user input on child widgets, except `clear_button` which is disabled.
        """
        self.clear_button.hide()
        self.calculate_button.show()
        if self.symbol_widget is not None:
            self.symbol_widget.layout().itemAt(1).widget().setEnabled(True)
            self.symbol_widget.layout().itemAt(1).widget().clear()
        for control in self.control_widgets:
            if self.control_widgets[control] is not None:
                if type(self.control_widgets[control].layout().itemAt(1).widget()) != QtWidgets.QRadioButton:
                    self.control_widgets[control].layout().itemAt(
                        1).widget().clear()
                self.control_widgets[control].layout().itemAt(
                    2).widget().setEnabled(True)
                self.control_widgets[control].layout().itemAt(
                    2).widget().setCheckState(QtCore.Qt.Unchecked)
                self.control_widgets[control].layout().itemAt(
                    1).widget().setEnabled(False)
                # NOTE: have to set Widget1 to disabled last since the signal on the Widget2 is connected
                # through CheckState to the enabled stated of Widget1, i.e. flipping the CheckState on
                # Widget2 switches the enabled state of Widget1.

    def fire(self) -> None:
        """
        Disables user input on child widgets, except `clear_button` which is enabled.
        """
        self.clear_button.show()
        self.calculate_button.hide()
        if self.symbol_widget is not None:
            self.symbol_widget.layout().itemAt(1).widget().setEnabled(False)
        for control in self.control_widgets:
            if self.control_widgets[control] is not None:
                self.control_widgets[control].layout().itemAt(
                    1).widget().setEnabled(False)
                self.control_widgets[control].layout().itemAt(
                    2).widget().setEnabled(False)


class TableWidget(QtWidgets.QWidget):
    """
    Base class for other more complex widgets to inherit. An instance of `scrilla.gui.widgets.TableWidget` embeds its child widgets in its own layout, which is an instance of``PySide6.QtWidgetQVBoxLayout``. This class provides access to the following widgets: a ``PySide6.QtWidgets.QLabel`` for the title widget, an error message widget and a ``PySide6.QtWidgets.QTableWidget`` for results display. 

    Parameters
    ----------
    1. **layer**: ``str``
        Stylesheet property attached to widget.
    2. **widget_title**: ``str``
        *Optional*. Defaults to "Table Result". Title of the widget.

    Attributes
    ----------
    1. **title**: ``PySide6.QtWidget.QLabel``
    2.. **table**: ``PySide6.QtWidget.QTableWidget``

    """

    def __init__(self, layer, widget_title: str = "Table Result"):
        super().__init__()
        self.layer = layer
        self._init_widgets(widget_title)
        self._arrange_widgets()
        self._stage_widgets()

    def _init_widgets(self, widget_title: str) -> None:
        """Creates child widgets and their layouts"""
        self.title_container = factories.layout_factory(
            layout='horizontal-box')
        self.title = factories.atomic_widget_factory(
            component='heading', title=widget_title)
        self.download_button = factories.atomic_widget_factory(
            component='download-button', title=None)
        self.table = factories.atomic_widget_factory(
            component='table', title=None)
        self.setLayout(QtWidgets.QVBoxLayout())

    def _arrange_widgets(self) -> None:
        factories.set_policy_on_widget_list([self, self.title, self.title_container],
                                            QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding,
                                                                  QtWidgets.QSizePolicy.Minimum))
        self.title_container.layout().addWidget(self.title)
        self.title_container.layout().addWidget(self.download_button)
        self.layout().addWidget(self.title_container)
        self.layout().addWidget(self.table, 1)

    def _stage_widgets(self) -> None:
        self.table.hide()
        self.download_button.hide()
        self.download_button.clicked.connect(self.show_file_dialog)

    def init_table(self, rows: List[str], columns: List[str]) -> None:
        """
        Initializes `table` for display. Number of rows and columns is determined by the length of the passed in lists.

        Parameters
        ----------
        1. **rows**: ``List[str]``
            List containing the row headers.
        2. **columns**: ``List[str]``
            List containing the column headers.
        """
        self.table.setRowCount(len(rows))
        self.table.setColumnCount(len(columns))
        self.table.setHorizontalHeaderLabels(columns)
        self.table.setVerticalHeaderLabels(rows)

    def show_table(self):
        self.table.resizeColumnsToContents()
        self.table.show()
        self.download_button.show()

    @QtCore.Slot()
    def show_file_dialog(self) -> None:
        file_path = factories.atomic_widget_factory(
            component='save-dialog', title=f'(*.{settings.FILE_EXT})')
        file_path.selectFile(f'table.{settings.FILE_EXT}')
        filename = None
        if file_path.exec_() == QtWidgets.QDialog.Accepted:
            filename = file_path.selectedFiles()
        if filename is not None and len(filename) > 0:
            if settings.FILE_EXT == 'json':
                utilities.download_table_to_json(self.table, filename[0])


class GraphWidget(QtWidgets.QWidget):
    """
    Base class for other more complex widgets to inherit. An instance of `scrilla.gui.widgets.GraphWidget` embeds its child widgets in its own layout, which is an instance of``PySide6.QtWidgetQVBoxLayout``. This class provides access to the following widgets: a ``PySide6.QtWidgets.QLabel`` for the title widget, an error message widget and a ``PySide6.QtWidgets.QTableWidget`` for results display. 

    The graph to be displayed should be stored in the `scrilla.settings.TEMP_DIR` directory before calling the `show_pixmap` method on this class. It will load the graph from file and convert it into a ``PySide6.QtGui.QPixMap``.

    Parameters
    ----------
    1. **tmp_graph_key**: ``str``
        The key of the file in the `scrilla.settings.TEMP_DIR` used to store the image of the graph.
    2. **widget_title**: ``str``
        *Optional*. Defaults to "Graph Results". Title of the widget.

    Attributes
    ----------
    1. **tmp_graph_key**: ``str``
    2. **title**: ``PySide6.QtWidget.QLabel``
    3. **figure**: ``PySide6.QtWidget.QLabel``
    """

    def __init__(self, tmp_graph_key: str, layer: str, widget_title: str = "Graph Results"):
        super().__init__()
        self.layer = layer
        self.tmp_graph_key = tmp_graph_key
        self._init_widgets(widget_title)
        self._arrange_widgets()
        self._stage_widgets()

    def _init_widgets(self, widget_title: str) -> None:
        self.title_container = factories.layout_factory(
            layout='horizontal-box')
        self.title = factories.atomic_widget_factory(
            component='heading', title=widget_title)
        self.download_button = factories.atomic_widget_factory(
            component='download-button', title=None)
        self.figure = factories.atomic_widget_factory(
            component='figure', title=None)
        self.setLayout(QtWidgets.QVBoxLayout())

    def _arrange_widgets(self) -> None:
        factories.set_policy_on_widget_list([self, self.title, self.title_container],
                                            QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding,
                                                                  QtWidgets.QSizePolicy.Minimum))
        self.title_container.layout().addWidget(self.title)
        self.title_container.layout().addWidget(self.download_button)
        self.layout().addWidget(self.title_container)
        self.layout().addWidget(self.figure)

    def _stage_widgets(self) -> None:
        self.figure.hide()
        self.download_button.hide()
        self.download_button.clicked.connect(self.show_file_dialog)

    def set_pixmap(self) -> None:
        self.figure.setPixmap(utilities.generate_pixmap_from_temp(
            self.width(), self.height(), self.tmp_graph_key))
        self.figure.show()
        self.download_button.show()

    def clear(self):
        self.figure.hide()
        self.download_button.hide()

    @QtCore.Slot()
    def show_file_dialog(self) -> None:
        file_path = factories.atomic_widget_factory(
            component='save-dialog', title=f'(*.{settings.IMG_EXT})')
        file_path.selectFile(f'{self.tmp_graph_key}.{settings.IMG_EXT}')
        filename = None
        if file_path.exec_() == QtWidgets.QDialog.Accepted:
            filename = file_path.selectedFiles()
        if filename is not None and len(filename) > 0:
            utilities.download_tmp_to_file(self.tmp_graph_key, filename[0])


class CompositeWidget(QtWidgets.QWidget):
    """
    Constructor
    -----------
    1. **tmp_graph_key**: ``str``
    2. **widget_title**: ``str``
        *Optional*. Defaults to "Results". Title of the widget.
    3. **table_title**: ``Union[str,None]``
    4. **graph_title**: ``Union[str,None]``

    Attributes
    ----------
    1. **title**: ``PySide6.QtWidgets.QLabel``
    2. **table_widget**: ``scrilla.gui.widgets.TableWidget``
    3. **graph_widget**: ``scrilla.gui.widgets.GraphWidget``
    4. **tab_widget**: ``PySide6.QtWidget.QTabWidget``

    """

    def __init__(self, tmp_graph_key: str, layer: str, widget_title: str = "Results",
                 table_title: Union[str, None] = None,
                 graph_title: Union[str, None] = None):
        super().__init__()
        self.setObjectName(layer)
        self._init_widgets(widget_title=widget_title,
                           tmp_graph_key=tmp_graph_key)
        self._arrange_widgets(graph_title=graph_title,
                              table_title=table_title)

    def _init_widgets(self, widget_title: str, tmp_graph_key: str) -> None:
        """Creates child widgets and their layouts"""
        self.title = factories.atomic_widget_factory(
            component='subtitle', title=widget_title)

        self.table_widget = TableWidget(layer=self.objectName())
        self.graph_widget = GraphWidget(
            tmp_graph_key=tmp_graph_key, layer=self.objectName())

        self.tab_widget = QtWidgets.QTabWidget()

        self.setLayout(QtWidgets.QVBoxLayout())

    def _arrange_widgets(self, graph_title: str, table_title: str) -> None:
        self.title.setSizePolicy(QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding,
                                                       QtWidgets.QSizePolicy.Minimum))
        self.tab_widget.setSizePolicy(QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding,
                                                            QtWidgets.QSizePolicy.Expanding))
        self.title.setAlignment(QtCore.Qt.AlignLeft)

        self.tab_widget.addTab(self.table_widget, table_title)
        self.tab_widget.addTab(self.graph_widget, graph_title)

        self.layout().addWidget(self.title)
        self.layout().addWidget(self.tab_widget)



import datetime
from scrilla.static import keys

HELP_MSG = [
    "A financial application for optimizing portfolio allocations, calculating financial statistics and generating graphical plots. This library requires API keys from Alpha Vantage (https://www.alphavantage.co), Quandl (https://www.quandl.com/) and IEX (https://iexcloud.io/) to hydrate with data. These keys should be stored in environment variables named ALPHA_VANTAGE_KEY, QUANDL_KEY and IEX_KEY.",

    "See documentation for more information on configuration and usage: https://chinchalinchin.github.io/scrilla/."
]

SYNTAX = "scrilla <command> <tickers/symbols> <options>"

FUNC_DICT = {
    "asset_type": {
        'name': 'Asset Type',
        'values': ["asset", "a"],
        'args': None,
        'description': "Outputs the asset type for the supplied symbol.",
        'tickers': True,
    },
    "cvar": {
        'name': 'Conditional Value At Risk',
        'values': ["cvar", "cv"],
        'args': ['probability', 'expiry', 'start_date', 'end_date', 'save_file', 'suppress_output', 'json', keys.keys['ESTIMATION']['MOMENT'], keys.keys['ESTIMATION']['PERCENT'], keys.keys['ESTIMATION']['LIKE']],
        'description': "Calculates the conditional value at risk, i.e. E(St | St < Sv) where Sv -> Prob(St<Sv) = `prob` , for the list of inputted ticker symbols. 'expiry' and 'prob' are required arguments for this function. Note: 'expiry' is measured in years and is different from the `start` and `end` dates. `start` and `end` are used to calibrate the model to a historical sample and `expiry` is used as the time horizon over which the value at risk is calculated into the future.",
        'tickers': True,
    },
    "var": {
        'name': 'Value At Risk',
        'values': ["var", "v"],
        'args': ['probability', 'expiry', 'start_date', 'end_date', 'save_file', 'suppress_output', 'json', keys.keys['ESTIMATION']['MOMENT'], keys.keys['ESTIMATION']['PERCENT'], keys.keys['ESTIMATION']['LIKE']],
        'description': "Calculates the value at risk, i.e. for a given p, the Sv such that Prob(St<Sv) = p. 'expiry' and 'prob' are required arguments for this function. Note: 'expiry' is measured in years and is different from the `start` and `end` dates. `start` and `end` are used to calibrate the model to a historical sample and `expiry` is used as the time horizon over which the value at risk is calculated into the future.",
        'tickers': True,
    },
    "capm_equity_cost": {
        'name': 'Capital Asset Pricing Model Cost of Equity',
        'values': ["capm-equity", "capm-e"],
        'args': ['start_date', 'end_date', 'save_file', 'suppress_output', 'json', keys.keys['ESTIMATION']['MOMENT'], keys.keys['ESTIMATION']['PERCENT'], keys.keys['ESTIMATION']['LIKE']],
        'description': "Computes the cost of equity according to CAPM for the supplied list of tickers. If no start or end dates are specified, calculations default to the last 100 days of prices. The environment variable MARKET_PROXY defines which ticker serves as a proxy for the market as whole.",
        'tickers': True,
    },
    "capm_beta": {
        'name': 'Capital Asset Pricing Model Beta',
        'values': ["capm-beta", "capm-b"],
        'args': ['start_date', 'end_date', 'save_file', 'suppress_output', 'json', keys.keys['ESTIMATION']['MOMENT'], keys.keys['ESTIMATION']['PERCENT'], keys.keys['ESTIMATION']['LIKE']],
        'description': "Computes the market beta according to CAPM for the supplied list of tickers. If no start or end dates are specified, calculations default to the last 100 days of prices. The environment variable MARKET_PROXY defines which ticker serves as a proxy for the market as whole.",
        'tickers': True,
    },
    "clear_cache": {
        'name': 'Clear Cache',
        'values': ["clear-cache", "cc"],
        'args': None,
        'description': "Clears the _installation_directiory_/data/cache/ directory of all data.",
        'tickers': False,
    },
    "clear_static": {
        'name': 'Clear Static',
        'values': ["clear-static", "cs"],
        'args': None,
        'description': "Clears the _installation_directory_/data/static/ directory of all data. Not recommended unless necessary; Static data takes a long time to reload.",
        'tickers': False,
    },
    "clear_common": {
        'name': 'Clear Common',
        'values': ["clear-common", "cc"],
        'args': None,
        'description': "Clears the _installation_directory_/data/common/, which includes API keys stored through the command line and ticker saved to the user watchlist.",
        'tickers': False,
    },
    "close": {
        'name': 'Last Closing Price',
        'values': ["close", "cl"],
        'args': None,
        'description': "Return latest closing value for the supplied list of symbols (equity or crypto).",
        'tickers': True,
    },
    "correlation": {
        'name': 'Correlation Matrix',
        'values': ["correlation", "cor"],
        'args': ['start_date', 'end_date', 'save_file', 'suppress_output', 'json', keys.keys['ESTIMATION']['MOMENT'], keys.keys['ESTIMATION']['PERCENT'], keys.keys['ESTIMATION']['LIKE']],
        'description': "Calculate pair-wise correlation for the supplied list of ticker symbols. If no start or end dates are specified, calculations default to the last 100 days of prices.",
        'tickers': True,
    },
    "correlation_time_series": {
        'name': 'Correlation Time Series',
        'values': ["correlations", "cors"],
        'args': ['start_date', 'end_date', 'save_file', 'suppress_output', 'json', keys.keys['ESTIMATION']['MOMENT'], keys.keys['ESTIMATION']['PERCENT'], keys.keys['ESTIMATION']['LIKE']],
        'description': "Calculate correlation time series for a pair of tickers over a specified date range. If no start or end dates are specified, the default analysis period of 100 days is applied.",
        'tickers': True,
    },
    "discount_dividend": {
        'name': 'Discount Dividend Model',
        'values': ["discount-dividend-model", "ddm"],
        'args': ['discount', 'save_file', 'suppress_output', 'json'],
        'description': "Extrapolates future dividend cashflows from historical dividend payments with linear regression and then uses that model to calculate the net present value of all future dividends. If no discount rate is specified, the calculations default to the asset's cost of equity as determined the by the CAPM model.",
        'tickers': True,
    },
    "dividends": {
        'name': 'Dividend History',
        'values': ["dividends", "divs"],
        'args': ['start_date', 'end_date', 'save_file', 'suppress_output', 'json'],
        'description': "Displays the price history over the specific date range. If no dates are provided, returns the entire dividend history.",
        'tickers': True,
    },
    "efficient_frontier": {
        'name': 'Portfolio Efficient Frontier',
        'values': ["efficient-frontier", "ef"],
        'args': ['start_date', 'end_date', 'investment', 'target', 'steps', 'save_file', 'suppress_output', 'json', keys.keys['ESTIMATION']['MOMENT'], keys.keys['ESTIMATION']['PERCENT'], keys.keys['ESTIMATION']['LIKE']],
        'description': "Generate a sample of the portfolio's efficient frontier for the supplied list of tickers. The efficient frontier algorithm will minimize a portfolio's volality for a given rate of return and then maximize its return, and then use these points to generate the rest of the frontier by taking increments along the line connecting the (risk,return) profile of the minimum volatility portfolio to the (risk, return) profile of the maximum return portfolio. The number of points calculated in the efficient frontier can be specifed as an integer with the -steps flag. If no -steps is provided, the value of the environment variable FRONTIER_STEPS will be used.",
        'tickers': True,
    },
    "help": {
        'name': 'Help Message',
        'values': ["help", "h"],
        'args': None,
        'description': "Print this help message.",
        'tickers': False,
    },
    "interest_history": {
        'name': 'Interest Rate History',
        'values': ["interest", "int"],
        'args': ['start_date', 'end_date', 'save_file', 'suppress_output', 'json'],
        'description': "Prints the interest histories for each inputted maturity over the specified date range. If no date range is given, price histories will default to the last 100 days. See `scrilla -yield` for list of maturities.",
        'tickers': True,
    },
    "list_watchlist": {
        'name': 'Display Watchlist',
        'values': ["watchlist", "w-ls"],
        'args': None,
        'description': "Lists the equity symbols currently saved to your watchlist.",
        'tickers': False,
    },
    "maximize_return": {
        'name': 'Maximize Portfolio Return',
        'values': ["max-return", "max"],
        'args': ['start_date', 'end_date', 'investment', 'target', 'save_file', 'suppress_output', 'json', keys.keys['ESTIMATION']['MOMENT'], keys.keys['ESTIMATION']['PERCENT'], keys.keys['ESTIMATION']['LIKE']],
        'description': "Maximize the return of the portfolio defined by the supplied list of ticker symbols. Returns an array representing the allocations to be made for each asset in a portfolio. If no start or end dates are specified, calculations default to the last 100 days of prices. You can specify an investment with the '-invest' flag, otherwise the result will be output in percentage terms. Note: This function will always allocate 100% to the asset with the highest return. It's a good way to check and see if there are bugs in the algorithm after changes.",
        'tickers': True,
    },
    "moving_averages": {
        'name': 'Moving Averages Series',
        'values': ["mov-averages", "mas"],
        'args': ['start_date', 'end_date', keys.keys['ESTIMATION']['MOMENT'], keys.keys['ESTIMATION']['PERCENT'], keys.keys['ESTIMATION']['LIKE']],
        'description': "Calculate the current moving averages. If no start or end dates are specified, calculations default to the last 100 days of prices.",
        'tickers': True,
    },
    "optimize_portfolio": {
        'name': 'Optimize Portfolio Variance',
        'values': ["optimize-portfolio", "opt"],
        'args': ['start_date', 'end_date', 'optimize_sharpe', 'investment', 'target', 'save_file', 'suppress_output', 'json', keys.keys['ESTIMATION']['MOMENT'], keys.keys['ESTIMATION']['PERCENT'], keys.keys['ESTIMATION']['LIKE']],
        'description': "Optimize a portfolio's risk return profile.  Returns an array representing the allocations to be made for each asset in a portfolio.",
        'tickers': True,
    },
    "optimize_portfolio_conditional_var": {
        'name': 'Optimize Portfolio Conditional Value At Risk',
        'values': ["optimize-cvar", "opt-cvar"],
        'args': ['start_date', 'end_date', 'investment', 'target', 'save_file', 'suppress_output', 'json', 'expiry', 'probability', keys.keys['ESTIMATION']['MOMENT'], keys.keys['ESTIMATION']['PERCENT'], keys.keys['ESTIMATION']['LIKE']],
        'description': "Optimizes the conditional value at risk, i.e. E(St | St < Sv) where Sv -> Prob(St<S0) = `prob` , for the portfolio defined by the list of inputted ticker symbols. Note: expiry is measured in years and is different from the start and end dates. Start and end dates are used to calibrate the model to a historical sample while expiry is used as the time horizon over which the value at risk is calculated into the future.",
        'tickers': True,
    },
    "plot_correlation": {
        'name': 'Plot Correlation Time Series',
        'values': ["plot-correlations", "plot-cors"],
        'args': ['start_date', 'end_date', 'save_file', keys.keys['ESTIMATION']['MOMENT'], keys.keys['ESTIMATION']['PERCENT'], keys.keys['ESTIMATION']['LIKE']],
        'description': "Generates a time series for the correlation of two ticker symbols over the specified date range.",
        'tickers': True,
    },
    "plot_dividends": {
        'name': 'Plot Discount Dividend Model',
        'values': ["plot-dividends", "plot-divs"],
        'args': ['save_file'],
        'description': "Generates a scatter plot graphic of the dividend history for the supplied ticker with a superimposed simple linear regression line. Note: this function only accepts one ticker at a time.",
        'tickers': True,
    },
    "plot_frontier": {
        'name': 'Plot Efficient Frontier',
        'values': ["plot-efficient-frontier", "plot-ef"],
        'args': ['start_date', 'end_date', 'save_file', 'steps', keys.keys['ESTIMATION']['MOMENT'], keys.keys['ESTIMATION']['PERCENT'], keys.keys['ESTIMATION']['LIKE']],
        'description': "Generates a scatter plot graphic of the portfolio\'s efficient frontier for the supplied list of tickers. The number of points calculated in the efficient frontier can be specifed as an integer with the -steps. If no -steps is provided, the value of the environment variable FRONTIER_STEPS will be used. If this value is not set, the function will default to a value of 5.",
        'tickers': True,
    },
    "plot_moving_averages": {
        'name': 'Plot Moving Averages Series',
        'values': ["plot-moving-averages", "plot-mas"],
        'args': ['start_date', 'end_date', 'save_file', keys.keys['ESTIMATION']['MOMENT'], keys.keys['ESTIMATION']['PERCENT'], keys.keys['ESTIMATION']['LIKE']],
        'description': "Generates a grouped bar chart of the moving averages for each equity in the supplied list of ticker symbols. If no start or end dates are specified, calculations default to the last 100 days of prices.",
        'tickers': True,
    },
    "plot_return_qq": {
        'name': 'Normality of Returns (QQ Plot)',
        'values': ["plot-return-qq", "plot-qq"],
        'args': ['start_date', 'end_date', 'save_file'],
        'description': "Generates a Q-Q Plot to graphically test the normality of returns for the inputted ticker symbol over the specified date range. If no start or date are specified, calculations default to the last 100 days of prices.",
        'tickers': True,
    },
    "plot_return_dist": {
        'name': 'Distribution of Returns',
        'values': ["plot-return-distribution", "plot-dist"],
        'args': ['start_date', 'end_date', 'save_file'],
        'description': 'Generates a histogram of the annualized daily return over the specified date range.',
        'tickers': True
    },
    "plot_risk_profile": {
        'name': 'Plot Risk Profile',
        'values': ["plot-risk-profile", "plot-rp"],
        'args': ['start_date', 'end_date', 'save_file', keys.keys['ESTIMATION']['MOMENT'], keys.keys['ESTIMATION']['PERCENT'], keys.keys['ESTIMATION']['LIKE']],
        'description': "Generates a scatter plot of the risk-return profile for symbol in the supplied list of ticker symbols. If no start or end dates are specified, calculations default to the last 100 days of prices.",
        'tickers': True,
    },
    "plot_yield_curve": {
        'name': 'Plot Yield Curve',
        'values': ["plot-yield-curve", "plot-yc"],
        'args': ['start_date', 'save_file'],
        'description': "Generates a plot of the latest United States Treasury Yield Curve. A yield curveo n a different date can be generated by specifying the date with an argument.",
        'tickers': False,
    },
    "price_history": {
        'name': 'Price History',
        'values': ["prices", "pr"],
        'args': ['start_date', 'end_date', 'save_file', 'suppress_output', 'json'],
        'description': "Prints the price histories for each inputted asset over the specified date range. If no date range is given, price histories will default to the last 100 days.",
        'tickers': True,
    },
    "purge": {
        'name': 'Purge All Data',
        'values': ["purge", "pu"],
        'args': None,
        'description': "Removes all files contained with the _installation_directory_/data/static/, _installation_directory_/data/cache/ and _installation_directory_/data/common/ directory, but retains the directories themselves.",
        'tickers': False,
    },
    "risk_free_rate": {
        'name': 'Risk Free Rate',
        'values': ["risk-free", "rf"],
        'args': None,
        'description': "Returns the current annualized US Treasury yield specified by the RISK_FREE environment variables. Allowable values for RISK_FREE environment variable: ONE_MONTH, TWO_MONTH, THREE_MONTH, SIX_MONTH, ONE_YEAR, TWO_YEAR, THREE_YEAR, FIVE_YEAR, SEVEN_YEAR, TEN_YEAR, TWENTY_YEAR, THIRTY_YEAR.",
        'tickers': False,
    },
    "risk_profile": {
        'name': 'Risk Profile',
        'values': ["risk-profile", "rp"],
        'args': ['start_date', 'end_date', 'save_file', 'suppress_output', 'json', keys.keys['ESTIMATION']['MOMENT'], keys.keys['ESTIMATION']['PERCENT'], keys.keys['ESTIMATION']['LIKE']],
        'description': "Calculate the risk-return profile for the supplied list of ticker symbols. If no start or end dates are specified, calculations default to the last 100 days of prices.",
        'tickers': True,
    },
    "screener": {
        'name': 'Watchlist Screener',
        'values': ["screen", "scr"],
        'args': ['discount', 'criteria'],
        'description': "Searchs equity spot prices that trade at a discount to the provided model. If no model is provided, the screener will default to the Discount Dividend Model. If no discount rate is provided, the screener will default to the cost of equity for a ticker calculated using the CAPM model.",
        'tickers': False,
    },
    "sharpe_ratio": {
        'name': 'Sharpe Ratio',
        'values': ["sharpe-ratio", "sr"],
        'args': ['start_date', 'end_date', 'save_file', 'json', keys.keys['ESTIMATION']['MOMENT'], keys.keys['ESTIMATION']['PERCENT'], keys.keys['ESTIMATION']['LIKE']],
        'description': "Computes the sharpe ratio for each tickers in the supplied list",
        'tickers': True,
    },
    "statistic": {
        'name': 'Last Reported Value of Financial Statistic',
        'values': ["stat", "s"],
        'args': ['save_file', 'suppress_output', 'json'],
        'description': "Retrieves the latest value for the supplied list of economic statistics. The available list of economic statistic can be found at https://www.quandl.com/data/FRED-Federal-Reserve-Economic-Data/documentation?anchor=growth; it is also stored in the minstallation_directory_/data/static/ directory of the application.",
        'tickers': True,
    },
    "statistic_history": {
        'name': 'Financial Statistic History',
        'values': ["stats", "ss"],
        'args': ['start_date', 'end_date', 'save_file', 'suppress_output', 'json'],
        'description': "Prints the statistic history for the supplied list of economic statistics.The available list of economic statistic can be found at https://www.quandl.com/data/FRED-Federal-Reserve-Economic-Data/documentation?anchor=growth; it is also stored in the _installation_directory_/data/static/ directory of the application.",
        'tickers': True,
    },
    "store": {
        'name': 'API Key Store',
        'values': ["store", "st"],
        'args': ['key', 'value'],
        'description': "Save API key to local _installation_directory_/data/common/ directory.",
        'tickers': False,
    },
    "version": {
        'name': 'Display Version',
        'values': ["version", "v"],
        'args': None,
        'description': "Display installed version of this library.",
        'tickers': False,
    },
    "watchlist": {
        'name': 'Stock Watchlist',
        'values': ["watch", "w"],
        'args': None,
        'description': "Saves the supplist list of tickers to your watchlist. These equity symbol are used by the screening algorithms when searching for stocks that trade at a discount.",
        'tickers': False,
    },
    "yield_curve": {
        'name': 'Latest Yield Curve',
        'values': ["yield-curve", "yc"],
        'args': ['start_date'],
        'description': "Displays the current United States Treasury Yield Curve.",
        'tickers': False,
    }
}
"""
A dictionary containing configuration information for function arguments. This dictionary is used at various points in the library, such as `scrilla.gui.widgets.functions` and `scrilla.util.helper.format_args`, to generate function interfaces. 
"""

GROUP_DICT = {
    'estimation_method': 'Estimation Method'
}
"""
"""

ARG_META_DICT = {
    'groups': ['estimation_method']
}

# Every argument has four ways of being inputted: short-dash-long, long-dash-long, short-dash-short, long-dash-short
ARG_DICT = {
    'start_date': {
        'name': 'Sample Start Date',
        'values': ['-start-date', '--start-date', '-start', '--start'],
        'description': 'Start date of historical sample',
        'default': '100 Trading Days Ago',
        'widget_type': 'date',
        'format': lambda s: datetime.datetime.strptime(s, '%Y-%m-%d').date(),
        'required': False,
        'syntax': '<YYYY-MM-DD>',
        'cli_only': False
    },
    'end_date': {
        'name': 'Sample End Date',
        'values': ['-end-date', '--end-date', '-end', '--end'],
        'description': 'End date of historical sample',
        'default': 'Last Trading Day',
        'widget_type': 'date',
        'format': lambda s: datetime.datetime.strptime(s, '%Y-%m-%d').date(),
        'required': False,
        'syntax': '<YYYY-MM-DD>',
        'cli_only': False
    },
    'target': {
        'name': 'Target Return',
        'values': ['-target-return', '--target-return', '-target', '--target'],
        'description': 'Constraint for target rate of return',
        'default': None,
        'widget_type': 'decimal',
        'format': float,
        'required': False,
        'syntax': '<value>',
        'cli_only': False
    },
    'discount': {
        'name': 'Discount Rate',
        'values': ['-discount-rate', '--discount-rate', '-discount', '--discount'],
        'description': 'Rate of return used to discount future cash flows to present',
        'default': 'Cost of equity',
        'widget_type': 'decimal',
        'format': float,
        'required': False,
        'syntax': '<value>',
        'cli_only': False
    },
    'investment': {
        'name': 'Total Investment',
        'values': ['-investment', '--investment', '-invest', '--invest'],
        'description': 'Total amount of money invested',
        'default': None,
        'widget_type': 'currency',
        'format': float,
        'required': False,
        'syntax': '<value>',
        'cli_only': False
    },
    'expiry': {
        'name': 'Expiration',
        'values': ['-expiry', '--expiry', '-exp', '--exp'],
        'description': 'Time until event expiration',
        'default': None,
        'widget_type': 'decimal',
        'format': float,
        'required': True,
        'syntax': '<value>',
        'cli_only': False
    },
    'probability': {
        'name': 'Loss Probability',
        'values': ['-probability', '--probability', '-prob', '--prob'],
        'description': 'Probability a loss will occur',
        'default': None,
        'widget_type': 'decimal',
        'format': float,
        'required': True,
        'syntax': '<value>',
        'cli_only': False
    },
    'steps': {
        'name': 'Efficient Frontier Data Points',
        'values': ['-frontier-steps', '--frontier-steps', '-steps', '--steps'],
        'description': 'Number of data points included in efficient frontier calculations',
        'default': 'FRONTIER_STEPS environment variable',
        'widget_type': 'integer',
        'format': int,
        'required': False,
        'syntax': '<value>',
        'cli_only': False
    },
    'criteria': {
        'name': 'Watchlist Screener Critia',
        'values': ['-criteria', '--criteria', '-crit', '--crit'],
        'description': 'Criteria used to sort saved watchlist',
        'default': None,
        'widget_type': 'select',
        'format': str,
        'required': False,
        'allowable': ['DDM', 'sharpe', 'volatility', 'return'],
        'syntax': '<value>',
        'cli_only': False
    },
    'save_file': {
        'name': 'Save File Location',
        'values': ['-save-file', '--save-file', '-save', '--save'],
        'description': 'Location where file will be saved',
        'default': None,
        'widget_type': None,
        'format': str,
        'required': False,
        'syntax': '<path>',
        'cli_only': True
    },
    'key': {
        'name': 'Key-Value Key',
        'values': ['-key', '--key', '-k', '--k'],
        'description': 'API Key that will be saved to __installation_dir__/data/common/',
        'default': None,
        'format': str,
        'widget_type': None,
        'required': True,
        'allowable': ["ALPHA_VANTAGE_KEY", "QUANDL_KEY", "IEX_KEY"],
        'syntax': '<value>',
        'cli_only': True
    },
    'value': {
        'name': 'Key-Value Value',
        'values': ['-value', '--value', '-v', '--v'],
        'description': 'API Key-Value that will be saved to __installation_dir__/data/common/',
        'default': None,
        'widget_type': None,
        'format': str,
        'required': True,
        'syntax': '<value>',
        'cli_only': True
    },
    'optimize_sharpe': {
        'name': 'Optimize Portfolio Sharpe',
        'values': ['-sharpe', '--sharpe', '-sh', '--sh'],
        'description': 'Flag to optimize portfolio Sharpe Ratio',
        'default': None,
        'widget_type': 'flag',
        'format': bool,
        'required': False,
        'syntax': None,
        'cli_only': False
    },
    'json': {
        'name': 'JSON Display',
        'values': ['-json', '--json', '-js', '--js'],
        'description': 'Print output to screen as JSON',
        'default': None,
        'widget_type': 'flag',
        'format': bool,
        'required': False,
        'syntax': None,
        'cli_only': True
    },
    'suppress_output': {
        'name': 'No Display',
        'values': ['-quiet', '--quiet', "-q", "--q"],
        'description': 'Suppress all console output',
        'default': None,
        'widget_type': 'flag',
        'format': bool,
        'required': False,
        'syntax': None,
        'cli_only': True
    },
    'moments': {
        'name': 'Moment Matching',
        'values': ['-moments', '--moments', '-mom', '--mom'],
        'description': 'Statistics are calculated using method of moment matching',
        'default': None,
        'widget_type': 'group',
        'format': 'group',
        'group': 'estimation_method',
        'required': False,
        'syntax': None,
        'cli_only': False
    },
    'percentiles': {
        'name': 'Percentile Matching',
        'values': ['-percentiles', '--percentiles', '-per', '--per'],
        'description': 'Statistics are calculated using method of percentile matching',
        'default': None,
        'widget_type': 'group',
        'format': 'group',
        'group': 'estimation_method',
        'required': False,
        'syntax': None,
        'cli_only': False
    },
    'likelihood': {
        'name': 'Maximum Likelihood',
        'values': ['-likelihood', '--likelihood', '-like', '--like'],
        'description': 'Statistics are calculated using maximum likelihood estimation',
        'default': None,
        'widget_type': 'group',
        'format': 'group',
        'group': 'estimation_method',
        'required': False,
        'syntax': None,
        'cli_only': False
    }
}
"""
A dictionary containing configuration information for application arguments. This dictionary is used at various points in the library, such as `scrilla.gui.widgets.functions` and `scrilla.util.helper`, to parse construct and parse input elements.

.. notes::
    * Every argument has four ways of being inputted: short-dash-long, long-dash-long, short-dash-short, long-dash-short, e.g. the following commands are all equivalent,
    ```
    scrilla risk-profile LMT GD LNT -json
    ```
    ```
    scrilla risk-profile LMT GD LNT --json
    ```
    ```
    scrilla risk-profile LMT GD LNT -js
    ```
    ```
    scrilla risk-profile LMT GD LNT --js
    ```
    * arguments with a format of `group` are mutually exclusively modes, similar to a radio button.
"""



# This file is part of scrilla: https://github.com/chinchalinchin/scrilla.

# scrilla is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3
# as published by the Free Software Foundation.

# scrilla is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with scrilla.  If not, see <https://www.gnu.org/licenses/>
# or <https://github.com/chinchalinchin/scrilla/blob/develop/main/LICENSE>.

"""
A module of statistical point estimators and likelihood functions.
"""

from os import path
from sys import path as sys_path
from typing import List
from numpy import log, sqrt, exp, inf
from scipy.stats import norm, multivariate_normal

if __name__ == "__main__":
    APP_DIR = path.dirname(path.dirname(path.abspath(__file__)))
    sys_path.append(APP_DIR)

from scrilla import settings, errors, cache
import scrilla.util.outputter as outputter

logger = outputter.Logger('scrilla.analysis.estimators', settings.LOG_LEVEL)
profile_cache = cache.ProfileCache()
correlation_cache = cache.CorrelationCache()


def univariate_normal_likelihood_function(params: list, data: list) -> float:
    """
    This function returns the likelihood of a vector of parameters being observed from a sample univariate data of normal data. It can be used as objective function input for `scipy.optimize`'s optimization methods. 

    Parameters
    ----------
    1. **x** : ``list``
        Array representing a vector of parameters , in this case the mean rate of return and volatility from a sample of data.
    2. **data** : ``list``
        A list of data that has been drawn from a univariate normal population.
    """
    likelihood = 0
    for point in data:
        likelihood += norm.logpdf(x=point, loc=params[0], scale=params[1])
    return likelihood


def bivariate_normal_likelihood_function(params: list, data: list) -> float:
    r"""
    Returns the likelihood of a vector of parameters being observed from a sample bivariate data of normal data. It can be used as objective function input for `scipy.optimize`'s optimization methods. 

    Parameters
    ----------
    1. **params** : ``list``
        Array representing a vector of parameters, in this case the mean rate of returns, voledatilities and covariance for a bivariate normal distribution. *Important*: The vector must be order: 1. params[0] = \\(\mu_x\\), params[1]=\\(\mu_y\\), params[2] = \\(\sigma_x\\), params[3] = \\(\sigma_y\\), params[4] = \\(\rho_{xy} \cdot \sigma_x \cdot \sigma_y\\). The matrix is parameterized in this manner in order to interface more easily with `scipy.optimize.minimize`.
    2. **data** : ``list``
        A list of data that has been drawn from a bivariate normal population. Must be formatted in the following manner: `[ [x1,y1], [x2,y2],...]`

    .. notes::
        * the covariance matrix of a bivariate normal distribution must be positive semi-definite (PSD) and non-singular. PSD can be checked with the [Slyvester Criterion](https://en.wikipedia.org/wiki/Sylvester%27s_criterion) or [Cauchy-Schwarz Inequality](https://en.wikipedia.org/wiki/Cauchy%E2%80%93Schwarz_inequality#Probability_theory). Since sample variance will always be positive, this reduces to checking the determinant of the covariance matrix is greater than 0. This function will return `numpy.inf` if the covariance matrix is singular or non-positive semi-definite.

    """
    mean = [params[0], params[1]]
    cov = [[params[2], params[4]], [params[4], params[3]]]

    determinant = params[2]*params[3] - params[4]**2
    if determinant == 0 or determinant < 0 or determinant < 0.00000001:
        return inf

    likelihood = 0
    for point in data:
        likelihood += multivariate_normal.logpdf(x=point, mean=mean, cov=cov)
    return likelihood


def sample_percentile(data: list, percentile: float):
    """
    Returns the observation in a sample data corresponding to the given percentile, i.e. the observation from a sorted sample where the percentage of the observations below that point is specified by the percentile. If the percentile falls between data points, the observation is smoothed based on the distance from the adjoining observations in the following manner,

    .. todo:: add latex here

    Parameters
    ----------
    1. **data** : ``list``
        Array representing the set of data whose percentile is to be calculated.
    2. **percentile**: ``float``
        The percentile corresponding to the desired observation.
    """
    data.sort()

    obs_number = (len(data) + 1)*percentile
    extrapolate = obs_number - int(obs_number)

    if extrapolate == 0:
        return data[int(obs_number)-1]
    if obs_number > len(data):
        return data[-1]
    first_index = int(obs_number) - 1
    second_index = first_index + 1
    weight = obs_number - int(obs_number)
    return (1-weight)*data[first_index] + weight*data[second_index]


def sample_correlation(x: list, y: list):
    """
    Returns the sample correlation calculated using the Pearson correlation coefficient estimator,

    .. todo:: Pearson coefficient formula here

    Parameters
    ----------
    1. **x**: ``list``
        The *x* sample of paired data (*x*, *y*). Must preserve order with **y**.
    2. **y**: ``list``
        The *y* sample of paired data (*x*, *y*). Must preserve order with **x**.

    Raises 
    ------
    1. `scrilla.errors.SampleSizeError` :
        If the sample sizes do not meet the requirements for estimation, this error will be thrown.
    2. **ValueError** :
        If the denominator of the correlation coefficient becomes too small for floating point arithmetic, this error is thrown.

    .. todos ::
        * Possibly wrap the correlation coefficient numerator and denominator in `Decimal` class before calculation to bypass the **ValueError** that occurs in some samples where the denominator is too small for the arithmetic to detect.
    """
    if len(x) != len(y):
        raise errors.SampleSizeError('Samples are not of comparable lengths')

    if len(x) in [0, 1]:
        raise errors.SampleSizeError(
            'Sample correlation cannot be computed for a sample size less than or equal to 1.')

    sumproduct, sum_x_squared, sum_x, sum_y, sum_y_squared = 0, 0, 0, 0, 0
    n = len(x)
    for i, item in enumerate(x):
        sumproduct += item*y[i]
        sum_x += item
        sum_x_squared += item**2
        sum_y += y[i]
        sum_y_squared += y[i]**2
    correl_num = ((n*sumproduct) - sum_x*sum_y)
    correl_den = sqrt((n*sum_x_squared-sum_x**2)*(n*sum_y_squared-sum_y**2))

    # LET'S DO SOME MATHEMATICS! (to get around division by zero!)
    #   Unfortunately, this only works when A and B > 0 because log
    #       of a negative number only exists in complex plane.
    #   1. correl = A/B
    #   2. log(correl) = log(A/B) = log(A) - log(B)
    #   3. exp(log(correl)) = exp(log(A/B))
    #   4. correl = exp(log(A/B))
    if correl_num > 0 and correl_den > 0:
        log_correl = log(correl_num) - log(correl_den)
        correlation = exp(log_correl)
    else:
        if correl_den != 0:
            correlation = correl_num / correl_den
        else:
            raise ValueError(
                'Denominator for correlation formula to small for division')

    return correlation


def recursive_rolling_correlation(correl_previous, new_x_observation, lost_x_obs,
                                  new_y_obs, lost_y_obs, n=settings.DEFAULT_ANALYSIS_PERIOD):

    pass


def sample_mean(x: list) -> float:
    r"""
    Returns the sample mean from a sample of data \\(\{x_1 , x_2, ... , x_n \}\\),

    $$ \bar{x} = \frac{\sum_{i=1}^{n} x_i}/{n} $$

    Parameters
    ----------
    1. **x**: ``list``
        List containing a sample of numerical data.

    Raises 
    ------
    1. `scrilla.errors.SampleSizeError`
        If ``len(x)==0``, this error will be thrown.
    """
    xbar, n = 0, len(x)

    if n == 0:
        raise errors.SampleSizeError(
            'Sample mean cannot be computed for a sample size of 0.')

    for i in x:
        xbar += i/n
    return xbar


def recursive_rolling_mean(xbar_previous, new_obs, lost_obs, n=settings.DEFAULT_ANALYSIS_PERIOD):
    xbar_next = xbar_previous + (new_obs - lost_obs)/n
    return xbar_next


def sample_variance(x: list):
    r"""
    Returns the sample variance from a sample of data \\(\{x_1 , x_2, ... , x_n \}\\),

    $$ s^2=\frac{\sum_{i=1}^{n} (x_i - \bar{x})^2}/{n-1} $$

    Parameters
    ----------
    1. **x**: ``list``
        List containing a sample of numerical data.

    Raises 
    ------
    1. `scrilla.errors.SampleSizeError`
    """

    mu, sigma, n = sample_mean(x=x), 0, len(x)

    if n in [0, 1]:
        raise errors.SampleSizeError(
            'Sample variance cannot be computed for a sample size less than or equal to 1.')

    for i in x:
        sigma += ((i-mu)**2)/(n-1)
    return sigma


def recursive_rolling_variance(var_previous, xbar_previous, new_obs, lost_obs, n=settings.DEFAULT_ANALYSIS_PERIOD):
    xbar_new = recursive_rolling_mean(xbar_previous=xbar_previous, new_obs=new_obs,
                                      lost_obs=lost_obs, n=n)
    var_new = var_previous + \
        (n/(n-1))*((new_obs**2 - lost_obs**2)/n + (xbar_previous**2-xbar_new**2))
    return var_new


def sample_covariance(x: list, y: list):
    """
    Parameters
    ----------
    1. **x**: ``list``
        The *x* sample of paired data (*x*, *y*). Must preserve order with **y**.
    2. **y**: ``list``
        The *y* sample of paired data (*x*, *y*). Must preserve order with **x**.

    Raises 
    ------
    1. `scrilla.errors.SampleSizeError`
        If ``len(x) != len(y)`` (samples of incomparable length) or ``len(x) in [0,1]`` (insufficient data/degrees of freedom), this error will be thrown.
    """

    if len(x) != len(y):
        raise errors.SampleSizeError('Samples are not of comparable length')

    if len(x) in [0, 1]:
        raise errors.SampleSizeError(
            'Sample correlation cannot be computed for a sample size less than or equal to 1.')

    n, covariance = len(x), 0

    x_mean, y_mean = sample_mean(x=x), sample_mean(x=y)

    for i, item in enumerate(x):
        covariance += (item - x_mean)*(y[i] - y_mean) / (n - 1)

    return covariance


def recursive_rolling_covariance(covar_previous, new_x_obs, lost_x_obs, previous_x_bar,
                                 new_y_obs, lost_y_obs, previous_y_bar, n=settings.DEFAULT_ANALYSIS_PERIOD):
    new_sum_term = new_x_obs*new_y_obs - lost_x_obs*lost_y_obs
    xy_cross_term = previous_x_bar*(new_y_obs-lost_y_obs)
    yx_cross_term = previous_y_bar*(new_x_obs-lost_x_obs)
    perturbation = (new_x_obs-lost_x_obs)*(new_y_obs-lost_y_obs) / n
    numerator = new_sum_term - xy_cross_term - yx_cross_term - perturbation
    covar_new = covar_previous + numerator / (n-1)
    return covar_new


def simple_regression_beta(x: list, y: list):
    """
    Parameters
    ----------
    1. **x**: ``list``
        The *x* sample of paired data (*x*, *y*). Must preserve order with **y**.
    2. **y**: ``list``
        The *y* sample of paired data (*x*, *y*). Must preserve order with **x**.

    Raises 
    ------
    1. `scrilla.errors.statistics.SampleSizeError`
        If ``len(x) != len(y)`` (samples of incomparable length) or ``len(x) < 3`` (insufficient data/degrees of freedom), this error will be thrown.
    """

    if len(x) != len(y):
        raise errors.SampleSizeError(f'len(x) = {len(x)} != len(y) = {len(y)}')
    if len(x) < 3:
        raise errors.SampleSizeError(
            f'Sample size of {len(x)} is less than the necessary degrees of freedom (n > 2) for regression estimation.')

    correl = sample_correlation(x=x, y=y)
    vol_x = sqrt(sample_variance(x=x))
    vol_y = sqrt(sample_variance(x=y))

    beta = correl * vol_y / vol_x
    return beta


def simple_regression_alpha(x: list, y: list):
    """
    Parameters
    ----------
    1. **x**: ``list``
        The *x* sample of paired data (*x*, *y*). Must preserve order with **y**.
    2. **y**: ``list``
        The *y* sample of paired data (*x*, *y*). Must preserve order with **x**.

    Raises 
    ------
    1. `scrilla.errors.SampleSizeError`
        If ``len(x) != len(y)`` (samples of incomparable length) or ``len(x) < 3`` (insufficient data/degrees of freedom), this error will be thrown.
    """

    if len(x) != len(y):
        raise errors.SampleSizeError(
            f'len(x) == {len(x)} != len(y) == {len(y)}')

    if len(x) < 3:
        raise errors.SampleSizeError(
            f'Sample size of {len(x)} is less than the necessary degrees of freedom (n > 2) for regression estimation.')

    y_mean, x_mean = sample_mean(y), sample_mean(x)

    alpha = y_mean - simple_regression_beta(x=x, y=y)*x_mean
    return alpha


def qq_series_for_sample(sample: list) -> List[list]:
    """
    Calculates the QQ series for a sample of data, i.e. the set defined by the ordered pair of sample percentiles and theoretical normal percentiles. A sample's normality can be assessed by how linear the result graph is.

    Parameters
    ----------
    1. **sample**: ``list``
        A sample of numerical data.
    """
    qq_series = []
    n = len(sample)
    for i in range(len(sample)):
        percentile = (i + 0.5)/n
        percentile_sample = sample_percentile(
            data=sample, percentile=percentile)
        percentile_norm = norm.ppf(q=percentile)
        qq_series += [[percentile_norm, percentile_sample]]

    return qq_series
